import pandas as pd
import glob
import subprocess
import os
import celescope
import scanpy as sc
import matplotlib
import matplotlib.pyplot as plt
import warnings
from celescope.tools import utils
from celescope.tools.step import s_common, Step

matplotlib.use("Agg")
warnings.filterwarnings("ignore")

TOOLS_DIR = os.path.dirname(celescope.tools.__file__)

CELL_TYPE_DICT = {
    "TCR": ["T_cells", "NKT_cells", "T cells", "NK T cells", "Tcells"],
    "BCR": [
        "Plasma_cells",
        "B_cells",
        "Mature_B_cell",
        "Plasma cells",
        "B cells",
        "Bcells",
    ],
}


class Mapping(Step):
    """
    ## Features

    - Output TSNE-plot of Assembled T/B Cells.

    ## Output
    - `06.mapping/{sample}_mapping.pdf` TSNE-plot of Assembled Cells.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.match_dir = args.match_dir
        if self.match_dir != "None":
            self.contig_file = glob.glob(
                f"{args.match_out}/matched_contig_annotations.csv"
            )[0]
        self.celltype_set = CELL_TYPE_DICT[self.seqtype]

        self.rds = glob.glob(f"{self.match_dir}/06.analysis/*.rds")
        self.assign_file = glob.glob(
            f"{self.match_dir}/06.analysis/*_auto_assign/*_auto_cluster_type.tsv"
        )
        self.h5ad = glob.glob(f"{self.match_dir}/06.analysis/*.h5ad")

    @staticmethod
    @utils.add_log
    def run_mapping(rds, contig, sample, outdir, assign):
        cmd = (
            f"Rscript {TOOLS_DIR}/VDJmapping.R "
            f"--rds {rds} "
            f"--VDJ {contig} "
            f"--sample {sample} "
            f"--outdir {outdir} "
            f"--assign_file {assign} "
            "2>&1 "
        )
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def process_rds(self):
        """Mapping result with matched scRNA."""
        Mapping.run_mapping(
            self.rds[0], self.contig_file, self.sample, self.outdir, self.assign_file[0]
        )

        meta = pd.read_csv(f"{self.outdir}/{self.sample}_meta.csv")
        metaTB = meta[meta["CellTypes"].isin(self.celltype_set)]
        mappedmeta = meta[meta["Class"] == "T/BCR"]
        mappedmetaTB = mappedmeta[mappedmeta["CellTypes"].isin(self.celltype_set)]

        self.add_metric(
            "Total Cell Number in Matched transcriptome", meta.shape[0], show=False
        )
        self.add_metric(
            "Cell Number Successfully Mapped to transcriptome",
            mappedmeta.shape[0],
            show=False,
        )
        self.add_metric(
            "T/B cell Number in Matched transcriptome", metaTB.shape[0], show=False
        )
        self.add_metric(
            "Cell Number Successfully Mapped to T/B cell in transcriptome",
            mappedmetaTB.shape[0],
            show=False,
        )

    @utils.add_log
    def process_h5ad(self):
        contig_file = pd.read_csv(self.contig_file)
        assembled_cells = set(contig_file.barcode)
        print(self.h5ad)
        h5ad = sc.read_h5ad(self.h5ad[0])
        h5ad.obs["status"] = "None"
        h5ad.obs.loc[h5ad.obs.index.isin(assembled_cells), "status"] = "T/BCR"

        sc.pl.tsne(
            h5ad,
            color="status",
            add_outline=True,
            legend_loc="on data",
            legend_fontsize=12,
            legend_fontoutline=2,
            frameon=False,
            title="Assembled Cells",
            palette=["C7", "C3"],
        )
        plt.savefig(f"{self.outdir}/{self.sample}_mapping.pdf", dpi=300)

    def run(self):
        if self.rds and self.assign_file:
            self.process_rds()
        if self.h5ad:
            self.process_h5ad()


def mapping(args):
    step_name = "mapping"
    mapping_obj = Mapping(args, step_name)
    mapping_obj.run()


def get_opts_mapping(parser, sub_program):
    parser.add_argument(
        "--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True
    )
    if sub_program:
        s_common(parser)
        parser.add_argument(
            "--match_dir", help="scRNA-seq match directory", required=True
        )
        parser.add_argument(
            "--match_out", help="assemble result from match directory", required=True
        )
    return parser



import math
from collections import defaultdict
import operator
import functools

import plotly
import plotly.express as px
import plotly.graph_objects as go

from celescope.tools import utils

PLOTLY_CONFIG = {
    "displayModeBar": True,
    "staticPlot": False,
    "showAxisDragHandles": False,
    "modeBarButtons": [["toImage", "resetScale2d"]],
    "scrollZoom": False,
    "displaylogo": False,
}

COLORS = px.colors.qualitative.Plotly + px.colors.qualitative.Alphabet

LAYOUT = {
    "height": 313,
    "width": 400,
    "margin": {
        "l": 45,
        "r": 35,
        "b": 30,
        "t": 30,
    },
}


def _round_float(x: float) -> float:
    """Round a float to the nearest value which can be represented with 4 decimal digits.

    In order to avoid representing floats with full precision, we convert them
    to a lower precision string and then convert that back to a float for use by `json.dumps`
    which will typically then output many fewer digits if possible.
    """
    return float(f"{x:.4g}")


def round_floats_in_list(x):
    """Lower the precision for a whole bunch of floats."""
    return [_round_float(x) for x in x]


class Plotly_plot:
    def __init__(self, df):
        self._df = df
        self._fig = None

    def plotly_plot(self):
        return plotly.offline.plot(
            self._fig, include_plotlyjs=False, output_type="div", config=PLOTLY_CONFIG
        )

    def get_plotly_div(self):
        return self.plotly_plot()


class Tsne_plot(Plotly_plot):
    def __init__(self, df_tsne, feature_name, discrete=True):
        super().__init__(df_tsne)

        self.feature_name = feature_name
        self.discrete = discrete
        title_feature_name = feature_name[0].upper() + feature_name[1:]
        self.title = f"t-SNE plot Colored by {title_feature_name}"

        self._layout = {}
        self._dot_size = 4
        self._df["size"] = self._dot_size
        self._df["barcode_index"] = list(range(1, len(self._df) + 1))
        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "showline": False,
            "ticks": None,
            "zeroline": True,
            "zerolinecolor": "black",
            "zerolinewidth": 0.7,
        }

        self.scatter_config = {
            "data_frame": df_tsne,
            "title": self.title,
            "x": self._str_coord1,
            "y": self._str_coord2,
            "size_max": self._dot_size,
            "hover_data": {
                self._str_coord1: False,
                self._str_coord2: False,
                self.feature_name: True,
                "barcode_index": True,
                "size": False,
            },
            "size": "size",
            "opacity": 0.9,
            "color": self.feature_name,
            "color_discrete_sequence": COLORS,
            "color_continuous_scale": px.colors.sequential.Jet,
        }

    def set_color_scale(self, color_scale):
        self.scatter_config["color_continuous_scale"] = color_scale

    @utils.add_log
    def get_plotly_div(self):
        if self.discrete:
            self.discrete_tsne_plot()
        else:
            self.continuous_tsne_plot()
        self.update_fig()

        return self.plotly_plot()

    @utils.add_log
    def discrete_tsne_plot(self):
        sum_df = self._df.groupby([self.feature_name]).agg("count").iloc[:, 0]
        percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
        res_dict = defaultdict(int)
        res_list = []
        for cluster in sorted(self._df[self.feature_name].unique()):
            name = f"{cluster}({percent_df[cluster]}%)"
            res_dict[cluster] = name
            res_list.append(name)

        self._df[self.feature_name] = self._df[self.feature_name].map(res_dict)

        self._fig = px.scatter(
            **self.scatter_config, category_orders={self.feature_name: res_list}
        )

    @utils.add_log
    def continuous_tsne_plot(self):
        self._fig = px.scatter(
            **self.scatter_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(title_text=self._str_coord1, **self.axes_config)

        self._fig.update_yaxes(title_text=self._str_coord2, **self.axes_config)

        self._fig.update_layout(
            self._layout,
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )


class Tsne_dropdown_plot(Plotly_plot):
    def __init__(self, df_tsne, name, feature_name_list, discrete=False):
        super().__init__(df_tsne)

        self.name = name
        self.feature_name_list = feature_name_list
        self.discrete = discrete
        self.title = f"t-SNE plot Colored by {self.name}"
        self.x_pos = 0.1
        self._layout = {}

        self._buttons = []
        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "showline": False,
            "ticks": None,
            "zeroline": True,
            "zerolinecolor": "black",
            "zerolinewidth": 0.7,
        }

        x_ = math.ceil(max(abs(self._df[self._str_coord1])))
        y_ = math.ceil(max(abs(self._df[self._str_coord2])))
        self.x_range = [-x_, x_]
        self.y_range = [-y_, y_]

    def get_plotly_div(self):
        if self.discrete:
            self.discrete_tsne_plot()
        else:
            self.continuous_tsne_plot()

        self.update_fig()

        return self.plotly_plot()

    def continuous_tsne_plot(self):
        self._fig = go.Figure()
        _num = len(self.feature_name_list)
        i = 0

        max_num = math.ceil(max([self._df[_].max() for _ in self.feature_name_list]))
        split_num = math.ceil(max_num / 5)
        tick_list = list(range(0, max_num, split_num))

        for feature_name in self.feature_name_list:
            df_tmp = self._df.loc[:, [self._str_coord1, self._str_coord2, feature_name]]
            self._fig.add_trace(
                go.Scattergl(
                    x=round_floats_in_list(df_tmp[self._str_coord1]),
                    y=round_floats_in_list(df_tmp[self._str_coord2]),
                    mode="markers",
                    name=feature_name[0].upper() + feature_name[1:],
                    showlegend=False,
                    marker=go.scattergl.Marker(
                        opacity=0.9,
                        size=3,
                        color=self._df[feature_name],
                        cmax=max_num,
                        cmin=0,
                        colorscale=[[0, "LightGrey"], [1, "Red"]],
                        colorbar=go.scattergl.marker.ColorBar(
                            tickmode="array",
                            tickvals=tick_list,
                            ticktext=tick_list,
                            title="",
                            titlefont={"size": 11},
                        ),
                    ),
                    textposition="top center",
                )
            )
            visible_list = [False] * _num
            visible_list[i] = True
            i += 1
            button = dict(
                args=[
                    {"visible": visible_list},
                ],
                label=feature_name,
                method="restyle",
            )
            self._buttons.append(button)

    def discrete_tsne_plot(self):
        all_figures = []
        num_all, last_num = 0, 0
        for feature_name in self.feature_name_list:
            num_all += len(self._df[feature_name].unique())

        for feature_name in self.feature_name_list:
            df_sub = self._df.loc[:, ["tSNE_1", "tSNE_2", feature_name]]
            num_feature = len(df_sub[feature_name].unique())

            sum_df = df_sub.groupby(feature_name).agg("count").iloc[:, 0]
            percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
            res_dict = defaultdict(int)
            res_list = []
            for cluster in sorted(df_sub[feature_name].unique()):
                name = f"{cluster}({percent_df[cluster]}%)"
                res_dict[cluster] = name
                res_list.append(name)

            df_sub[feature_name] = df_sub[feature_name].map(res_dict)
            df_sub["size"] = 4

            scatter_config = {
                "data_frame": df_sub,
                "x": "tSNE_1",
                "y": "tSNE_2",
                "size_max": 4,
                "hover_data": {
                    "tSNE_1": False,
                    "tSNE_2": False,
                    feature_name: True,
                    "size": False,
                },
                "size": "size",
                "opacity": 0.9,
                "color": feature_name,
                "color_discrete_sequence": COLORS,
                "color_continuous_scale": px.colors.sequential.Jet,
            }
            fig = px.scatter(**scatter_config, category_orders={feature_name: res_list})
            all_figures.append(fig)

            true_list = [True] * num_feature
            visible_list = [False] * num_all

            visible_list[last_num : num_feature + last_num] = true_list
            last_num += num_feature
            button = dict(
                args=[
                    {"visible": visible_list},
                ],
                label=feature_name,
                method="update",
            )
            self._buttons.append(button)

        self._fig = go.Figure(
            data=functools.reduce(operator.add, [_.data for _ in all_figures])
        )

    def update_fig(self):
        self._fig.update_xaxes(
            range=self.x_range, title_text=self._str_coord1, **self.axes_config
        )

        self._fig.update_yaxes(
            range=self.y_range, title_text=self._str_coord2, **self.axes_config
        )

        self._fig.update_layout(
            self._layout,
            updatemenus=[
                {
                    "buttons": self._buttons,
                    "direction": "down",
                    "pad": {"r": 8, "t": 8},
                    "showactive": True,
                    "x": 1.1,
                    "xanchor": "right",
                    "y": 1.2,
                    "yanchor": "top",
                }
            ],
            annotations=[
                {
                    "text": "TAG type:",
                    "showarrow": False,
                    "x": self.x_pos,
                    "xref": "paper",
                    "xanchor": "right",
                    "y": 1.15,
                    "yref": "paper",
                    "font": {"size": 11},
                    "align": "left",
                }
            ],
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )


# Add single tag plot
class Tsne_single_plot(Plotly_plot):
    def __init__(self, df_tsne, feature_name_list, analysis_dir):
        super().__init__(df_tsne)

        self.feature_name_list = feature_name_list
        self.analysis_dir = analysis_dir

        self._layout = {}

        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "showline": False,
            "ticks": None,
            "zeroline": True,
            "zerolinecolor": "black",
            "zerolinewidth": 0.7,
        }

        x_ = math.ceil(max(abs(self._df[self._str_coord1])))
        y_ = math.ceil(max(abs(self._df[self._str_coord2])))
        self.x_range = [-x_, x_]
        self.y_range = [-y_, y_]

    def get_plotly_div(self):
        max_num = math.ceil(max([self._df[_].max() for _ in self.feature_name_list]))
        split_num = math.ceil(max_num / 5)
        tick_list = list(range(0, max_num, split_num))

        for feature_name in self.feature_name_list:
            self._fig = go.Figure()
            name = feature_name[0].upper() + feature_name[1:]
            title = f"t-SNE plot Colored by {name}"
            df_tmp = self._df.loc[:, [self._str_coord1, self._str_coord2, feature_name]]
            df_tmp = df_tmp.loc[df_tmp[feature_name] != 0]
            self._fig.add_trace(
                go.Scatter(
                    x=round_floats_in_list(df_tmp[self._str_coord1]),
                    y=round_floats_in_list(df_tmp[self._str_coord2]),
                    mode="markers",
                    showlegend=False,
                    marker=go.scatter.Marker(
                        opacity=0.9,
                        size=3,
                        color=self._df[feature_name],
                        cmax=max_num,
                        cmin=0,
                        colorscale=[[0, "LightGrey"], [1, "Red"]],
                        colorbar=go.scatter.marker.ColorBar(
                            tickmode="array",
                            tickvals=tick_list,
                            ticktext=tick_list,
                            title="",
                            titlefont={"size": 11},
                        ),
                    ),
                    textposition="top center",
                )
            )

            self.update_fig()
            self._fig.update_layout(
                title={"text": title, "x": 0.5, "y": 0.95, "font": {"size": 15}}
            )
            self._fig.write_image(f"{self.analysis_dir}/{name}_tsne.png")

    def update_fig(self):
        self._fig.update_xaxes(
            range=self.x_range, title_text=self._str_coord1, **self.axes_config
        )

        self._fig.update_yaxes(
            range=self.y_range, title_text=self._str_coord2, **self.axes_config
        )

        self._fig.update_layout(
            self._layout, plot_bgcolor="#FFFFFF", hovermode="closest"
        )


class Bar_plot(Plotly_plot):
    def __init__(self, df_bar):
        super().__init__(df_bar)
        self.set_fig()

    def set_fig(self):
        self._fig = px.bar(
            x=[str(i) for i in list(self._df.head(10).ClonotypeID)],
            y=self._df.head(10).proportion.tolist(),
            labels={"x": "Clonotype ID", "y": "Fraction of Cells"},
            width=700,
            height=500,
        )
        self._fig.update_traces(
            marker_color="rgb(158,202,225)",
            marker_line_color="rgb(8,48,107)",
            marker_line_width=1.5,
            opacity=0.6,
        )
        self._fig.update_layout(
            title_text="Top 10 Clonotype Frequencies",
            title={"x": 0.5, "y": 0.9, "font": {"size": 20, "family": "San Serif"}},
            plot_bgcolor="#FFFFFF",
        )


class Pie_plot(Plotly_plot):
    def __init__(self, df_region):
        super().__init__(df_region)
        self.set_fig()

    def set_fig(self):
        layout = {
            "height": 300,
            "width": 400,
            "margin": {
                "l": 50,
                "r": 35,
                "b": 10,
                "t": 10,
            },
        }
        self._fig = px.pie(
            self._df,
            names="regions",
            values="values",
        )
        self._fig.update_traces(textposition="none")
        self._fig.update_layout(layout)


class Line_plot(Plotly_plot):
    def __init__(
        self,
        df_line,
        title=None,
        x_title=None,
        y_title=None,
        y_range=None,
        section=True,
    ):
        super().__init__(df_line)
        self.df_line = df_line
        self.title = title
        self.x_title = x_title
        self.y_title = y_title
        self.y_range = y_range
        self.section = section

        self.xaxes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "linecolor": "black",
            "showline": True,
            "ticks": None,
            "tickmode": "linear",
            "tick0": 0,
            "dtick": 0.5,
        }

        if self.section:
            self.yaxes_config = {
                "showgrid": True,
                "gridcolor": "#F5F5F5",
                "linecolor": "black",
                "showline": True,
                "ticks": None,
                "rangemode": "tozero",
            }
        else:
            self.yaxes_config = {
                "showgrid": True,
                "gridcolor": "#F5F5F5",
                "linecolor": "black",
                "showline": True,
                "ticks": None,
                "range": self.y_range,
            }

        self.line_config = {
            "data_frame": df_line,
            "title": self.title,
            "x": self.x_title,
            "y": self.y_title,
        }

        self.line_plot()
        self.update_fig()

    @utils.add_log
    def line_plot(self):
        self._fig = px.line(
            **self.line_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(**self.xaxes_config)

        self._fig.update_yaxes(**self.yaxes_config)

        self._fig.update_layout(
            LAYOUT,
            title={"x": 0.5, "y": 0.95, "font": {"size": 15}},
            yaxis_zeroline=True,
            showlegend=False,
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )



#!/bin/env python
# coding=utf8

import os
import pandas as pd
import plotly
import plotly.graph_objects as go
from celescope.tools.step import Step, s_common
from celescope.tools import utils


class Replace_tsne(Step):
    """
    ## Features
    - Replace rate in each cluster
    - Top replace genes in each cluster

    ## Output
    - `{sample}.rep_in_tsne.txt` Replace rate in each cluster.
    - `{sample}.rep_in_tsne_top10` Top 10 replace genes in each cluster.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # input files
        self.sample = args.sample
        self.tsnefile = args.tsne
        self.matfile = args.mat
        self.repfile = args.rep
        self.mincell = args.mincell
        self.topgene = args.topgene
        # output files
        self.outdot = os.path.join(self.outdir, self.sample + ".rep_in_tsne.txt")
        self.outtbl = os.path.join(self.outdir, self.sample + ".rep_in_tsne_top10.txt")

    @utils.add_log
    def run(self):
        # rep in cells in cluster
        self.dot_tsne(self.repfile, self.tsnefile, self.outdot)
        div_item = self.tsne_plot(self.outdot)
        # high rep gene in each cluster
        self.top_gene_cluster(
            self.matfile, self.tsnefile, self.outtbl, self.mincell, self.topgene
        )
        tbltxt = pd.read_csv(self.outtbl, header=0, sep="\t")
        tbldiv = self.tsne_table(tbltxt)

        # report
        self.report_prepare(div_item, tbldiv)
        self._clean_up()

    @utils.add_log
    def dot_tsne(self, repfile, tsnefile, outfile):
        cells = {}
        with open(repfile, "r") as f:
            for i in f:
                ii = i.strip().split()
                cells[ii[0]] = ii[1]

        outf = open(outfile, "w")
        outf.write("Cell\ttSNE_1\ttSNE_2\tCluster\tratio\n")
        with open(tsnefile, "r") as f:
            f.readline()
            for i in f:
                ii = i.strip().split()
                if ii[0] in cells:
                    outl = "\t".join(ii[0:4]) + "\t" + cells[ii[0]] + "\n"
                else:
                    outl = "\t".join(ii[0:4]) + "\t0" + "\n"
                outf.write(outl)
        outf.close()

    @utils.add_log
    def tsne_plot(self, txt):
        df = pd.read_table(txt)
        df.sort_values(by="ratio")
        newtitle = "t-SNE plot Colored by RNA Turn-over rate"

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=df["tSNE_1"],
                y=df["tSNE_2"],
                mode="markers",
                marker_opacity=0.9,
                marker_size=4,
                marker_color=df["ratio"],
                marker_colorscale="PuBu",
                marker_showscale=True,
            )
        )
        fig.update_layout(height=600, width=600, title_text=newtitle)
        fig.update_layout(plot_bgcolor="#FFFFFF")
        fig.update_xaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            title_text="t-SNE1",
        )
        fig.update_yaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            title_text="t-SNE2",
        )

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type="div")

        return div

    def tsne_table(self, txt):
        marker_gene_table = txt.to_html(
            escape=False,
            index=False,
            table_id="replacement_table_cluster",
            justify="center",
        )

        return marker_gene_table

    def file_stat(self, infile, clu):
        clus = list(set(clu.values()))
        cluster = {}
        for c in clus:
            cluster[c] = {}
        fn = open(infile, "r")
        fnh = fn.readline().strip().split()
        fnh.insert(0, "")
        for i in fn:
            ii = i.strip().split()
            for j in range(1, len(ii)):
                if ii[j] == "NA":
                    continue
                if fnh[j] not in clu:
                    continue
                if ii[0] not in cluster[clu[fnh[j]]]:
                    cluster[clu[fnh[j]]][ii[0]] = []
                cluster[clu[fnh[j]]][ii[0]].append(float(ii[j]))
        fn.close()
        return cluster

    def tsne_file(self, infile):
        clu = {}
        with open(infile) as f:
            f.readline()
            for i in f:
                ii = i.strip().split()
                clu[ii[0]] = ii[3]
        return clu

    @utils.add_log
    def top_gene_cluster(self, matrix, tsnefile, outfile, mincell=5, topgene=10):
        tsne = self.tsne_file(tsnefile)
        cluster = self.file_stat(matrix, tsne)

        w = open(outfile, "w")
        w.write("cluster\tgene\tTurn-over_rate\tcells\n")
        for c in cluster:
            tmp = {}
            for g in cluster[c]:
                gt = sum(cluster[c][g]) / len(cluster[c][g])
                tmp[g] = gt
            sorttmp = sorted(tmp.items(), key=lambda item: item[1], reverse=True)
            tmpn = 0
            for x in sorttmp:
                if len(cluster[c][x[0]]) < mincell:
                    continue
                tmpn += 1
                if tmpn > topgene:
                    break
                w.write(
                    "cluster"
                    + c
                    + "\t"
                    + x[0]
                    + "\t"
                    + str(x[1])
                    + "\t"
                    + str(len(cluster[c][x[0]]))
                    + "\n"
                )
        w.close()

    def report_prepare(self, outdiv, outable):
        self.add_data(replace_tsne=outdiv)
        self.add_data(replace_tsne_table=outable)


@utils.add_log
def replace_tsne(args):
    with Replace_tsne(args) as runner:
        runner.run()


def get_opts_replace_tsne(parser, sub_program):
    if sub_program:
        parser.add_argument(
            "--tsne", help="tsne file from analysis step", required=True
        )
        parser.add_argument(
            "--mat",
            help="matrix replacement file, from replacement step",
            required=True,
        )
        parser.add_argument(
            "--rep", help="cell replacement file, from replacement step", required=True
        )
        parser.add_argument(
            "--mincell",
            type=int,
            default=5,
            help="turn-over in at least cells, default 5",
        )
        parser.add_argument(
            "--topgene", type=int, default=10, help="show top N genes,default 10"
        )
        parser = s_common(parser)
    return parser



from celescope.tools import utils
from celescope.tools.mkref import Mkref, super_opts


class Mkref_rna(Mkref):
    """
    ## Features
    - Create a genome reference directory.

    ## Usage
    ```
    celescope utils mkgtf Homo_sapiens.GRCh38.99.gtf Homo_sapiens.GRCh38.99.filtered.gtf
    celescope rna mkref \\
    --genome_name Homo_sapiens_ensembl_99_filtered \\
    --fasta Homo_sapiens.GRCh38.dna.primary_assembly.fa \\
    --gtf Homo_sapiens.GRCh38.99.filtered.gtf
    ```

    ## Output

    - STAR genome index files

    - Genome config file
    ```
    $ cat celescope_genome.config
    [genome]
    genome_name = Homo_sapiens_ensembl_99
    genome_type = rna
    fasta = Homo_sapiens.GRCh38.dna.primary_assembly.fa
    gtf = Homo_sapiens.GRCh38.99.gtf
    ```
    """

    @utils.add_log
    def build_rna_star_index(self):
        cmd = (
            f"STAR \\\n"
            f"--runMode genomeGenerate \\\n"
            f"--runThreadN {self.thread} \\\n"
            f"--genomeDir ./ \\\n"
            f"--genomeFastaFiles {self.fasta} \\\n"
            f"--sjdbGTFfile {self.gtf} \\\n"
            f"--sjdbOverhang 100 \\\n"
        )
        if self.STAR_param:
            cmd += " " + self.STAR_param
        self.build_star_index.logger.info(cmd)
        self.debug_subprocess_call(cmd)

    @staticmethod
    def parse_genomeDir(genomeDir):
        return Mkref.parse_genomeDir(genomeDir, files=("gtf", "mt_gene_list"))

    @utils.add_log
    def run(self):
        super().run()
        self.build_star_index()


def mkref(args):
    genome_type = "rna"
    # files do not contain refflat because refflat is not input argument
    with Mkref_rna(genome_type, args, files=("gtf", "mt_gene_list")) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    super_opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--gtf",
            help="Required. Genome gtf file. Use absolute path or relative path to `genomeDir`.",
            required=True,
        )
        parser.add_argument(
            "--mt_gene_list",
            help="""Mitochondria gene list file. Use absolute path or relative path to `genomeDir`.
It is a plain text file with one gene per line. 
If not provided, will use `MT-` and `mt-` to determine mitochondria genes.""",
            default="None",
        )



#!/bin/env python
# coding=utf8

import os
import sys
import subprocess
import pandas as pd
import pysam
from celescope.tools.step import Step, s_common
from celescope.tools import utils

toolsdir = os.path.dirname(__file__)


class Replacement(Step):
    """
    Features
    - Computes the replacement rates in each cell and gene.
    - Boxplots for rates distribution.

    Output
    - `{sample}.new_matrix.tsv.gz` New RNA matrix.
    - `{sample}.old_matrix.tsv.gz` Old RNA matrix.
    - `{sample}.fraction_of_newRNA_per_cell.txt` Fraction of new RNA of each cell.
    - `{sample}.fraction_of_newRNA_per_gene.txt` Fraction of new RNA of each gene.
    - `{sample}.fraction_of_newRNA_matrix.txt` Fraction of new RNA of each cell and gene.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # input files
        self.outdir = args.outdir
        self.sample = args.sample
        self.bam_file = args.bam
        self.snp_file = args.bg
        self.bg_cov = args.bg_cov
        self.snp_threshold = args.snp_threshold

        # output files
        self.outmat = os.path.join(self.outdir, self.sample + ".TC_matrix.tsv")
        self.outpre = os.path.join(self.outdir, self.sample)

    @utils.add_log
    def run(self):
        # get backgroud snp
        bg = self.background_snp(self.snp_file, self.bg_cov, self.snp_threshold)
        # get reads with TC
        outframe = self.extract_dem(self.bam_file, bg)
        # run_R
        self.generate_TC_matrix(outframe, self.outmat)

        # split to New and Old Matrix
        new_mat = self.outpre + ".new_matrix.tsv"
        old_mat = self.outpre + ".old_matrix.tsv"
        con_mat = self.outpre + ".NvsO_matrix.tsv"
        self.split_matrix(self.outmat, self.outpre)

        # replacement stat
        self.replacment_stat(con_mat, self.outpre)
        # plot
        div_item = self.replacment_plot(self.outpre)

        # report
        self.report_prepare(div_item)
        self._clean_up()

        # clean
        cmd = ["rm", self.outmat]
        self.run_cmd(cmd)
        cmd = ["rm", con_mat]
        self.run_cmd(cmd)
        cmd = ["gzip", new_mat]
        self.run_cmd(cmd)
        cmd = ["gzip", old_mat]
        self.run_cmd(cmd)

    def run_cmd(self, cmd):
        subprocess.call(" ".join(cmd), shell=True)

    @utils.add_log
    def extract_dem(self, bam, bg):
        bamfile = pysam.AlignmentFile(bam, "rb")
        countdict = {}
        for read in bamfile.fetch():
            try:
                chro = read.reference_name
                cb = read.get_tag("CB")
                ub = read.get_tag("UB")
                if not read.has_tag("GN"):
                    continue
                gene = read.get_tag("GN")

                if read.get_tag("ST") == "+":
                    stag = read.get_tag("TL")
                else:
                    stag = read.get_tag("AL")
                if len(stag) == 1 and stag[0] == 0:
                    gene += "--T"
                else:
                    fcount = 0
                    for si in range(0, len(stag)):
                        pos = chro + "_" + str(stag[si])
                        if pos in bg:
                            fcount += 1
                    if fcount == len(stag):
                        gene += "--T"
                    else:
                        gene += "--C"

                readinfo = "\t".join([gene, cb, ub])
                if readinfo not in countdict:
                    countdict[readinfo] = 1

            except (ValueError, KeyError):
                continue
        bamfile.close()

        out1 = []
        for rid in countdict:
            checkt = rid.split("--T")
            if len(checkt) == 2:
                pairc = rid.replace("--T", "--C")
                if pairc not in countdict:
                    out1.append(rid.split("\t"))
            else:
                out1.append(rid.split("\t"))

        outframe = pd.DataFrame(out1)
        outframe.columns = ["geneID", "Barcode", "UMI"]
        out1 = []
        countdict = {}

        return outframe

    @utils.add_log
    def background_snp(self, bgfiles, cov=1, snp_threshold=0.5):
        ## dict.update()
        outdict = {}
        bgs = bgfiles.strip().split(",")
        for bgfile in bgs:
            if bgfile.endswith(".csv"):
                df = pd.read_csv(bgfile, index_col=0, dtype={"chrom": str})
                df = df[df["convs"] >= cov]
                if "pos" in df.columns:
                    df["chrpos"] = df["chrom"] + "_" + df["pos"].astype(str)
                else:  # compatible with previous version
                    df["chrpos"] = df["chrom"] + "_" + df["pos2"].astype(str)
                df1 = df[["chrpos", "posratio"]]
                df1.set_index("chrpos", inplace=True)
                outdict.update(df1.to_dict(orient="index"))
            elif bgfile.endswith(".vcf"):
                bcf_in = pysam.VariantFile(bgfile)
                for rec in bcf_in.fetch():
                    try:
                        chrom, pos = rec.chrom, rec.pos
                        chr_pos = chrom + "_" + str(pos - 1)
                        outdict[chr_pos] = 1
                    except (ValueError, KeyError):
                        continue
                bcf_in.close()
            elif bgfile.upper() == "SELF":
                selfbg = os.path.splitext(self.bam_file)[0] + ".csv"
                df = pd.read_csv(selfbg, index_col=0, dtype={"chrom": str})
                df = df[df["convs"] >= cov]
                df = df[df["posratio"] >= snp_threshold]
                if "pos" in df.columns:
                    df["chrpos"] = df["chrom"] + "_" + df["pos"].astype(str)
                else:  # compatible with previous version
                    df["chrpos"] = df["chrom"] + "_" + df["pos2"].astype(str)
                df1 = df[["chrpos", "posratio"]]
                df1.set_index("chrpos", inplace=True)
                outdict.update(df1.to_dict(orient="index"))
                continue
            else:
                try:
                    sys.exit(1)
                except SystemExit:
                    print(
                        "Background snp file format cannot be recognized! Only csv or vcf format."
                    )
                finally:
                    print(
                        "Background snp file format cannot be recognized! Only csv or vcf format."
                    )
        return outdict

    @utils.add_log
    def generate_TC_matrix(self, read, outmat):
        table = (
            read.pivot_table(
                index="geneID", columns="Barcode", values="UMI", aggfunc=len
            )
            .fillna(0)
            .astype(int)
        )
        table.index.name = ""
        table.to_csv(outmat, sep="\t")

    @utils.add_log
    def split_matrix(self, mat, outpre):
        outnew = open(outpre + ".new_matrix.tsv", "w")
        outold = open(outpre + ".old_matrix.tsv", "w")
        con_mat = open(outpre + ".NvsO_matrix.tsv", "w")
        infile = open(mat, "r")

        tmph = infile.readline().strip().split()
        fill_na = ["0"] * len(tmph)
        tmph.insert(0, "")
        outnew.write("\t".join(tmph) + "\n")
        outold.write("\t".join(tmph) + "\n")
        con_mat.write("\t".join(tmph) + "\n")

        genes = {}
        for i in infile:
            ii = i.strip().split()
            gt = ii[0].split("--")
            ii[0] = gt[0]
            if gt[0] not in genes:
                genes[gt[0]] = [0, [], []]
            if gt[1] == "C":
                genes[gt[0]][0] += 1
                genes[gt[0]][1] = ii[1:]
                outnew.write("\t".join(ii) + "\n")
            elif gt[1] == "T":
                genes[gt[0]][0] += 2
                genes[gt[0]][2] = ii[1:]
                outold.write("\t".join(ii) + "\n")

        for gi in genes:
            con_mat.write(gi)
            if genes[gi][0] == 3:
                for ci in range(len(genes[gi][1])):
                    con_mat.write("\t" + genes[gi][1][ci] + ":" + genes[gi][2][ci])
                con_mat.write("\n")
            elif genes[gi][0] == 1:
                outold.write(gi + "\t" + "\t".join(fill_na) + "\n")
                for ci in range(len(genes[gi][1])):
                    con_mat.write("\t" + genes[gi][1][ci] + ":" + "0")
                con_mat.write("\n")
            elif genes[gi][0] == 2:
                outnew.write(gi + "\t" + "\t".join(fill_na) + "\n")
                for ci in range(len(genes[gi][2])):
                    con_mat.write("\t" + "0" + ":" + genes[gi][2][ci])
                con_mat.write("\n")

        outnew.close()
        outold.close()
        con_mat.close()
        infile.close()

    @utils.add_log
    def replacment_stat(self, inmat, outpre, mincell=10, mingene=10):
        outcell = open(outpre + ".fraction_of_newRNA_per_cell.txt", "w")
        outgene = open(outpre + ".fraction_of_newRNA_per_gene.txt", "w")
        outmat1 = open(outpre + ".fraction_of_newRNA_matrix.txt", "w")

        cells = {}
        genes = {}
        mats = {}
        with open(inmat) as f:
            hh = f.readline().strip().split()
            hh.insert(0, "")
            outmat1.write("\t".join(hh) + "\n")
            for h in hh[1:]:
                cells[h] = [[], []]
            for i in f:
                ii = i.strip().split()
                mats[ii[0]] = []
                genes[ii[0]] = [[], []]
                for xi in range(1, len(ii)):
                    xx = [int(x) for x in ii[xi].split(":")]
                    if sum(xx) == 0:
                        tmpf = "NA"
                    else:
                        tmpf = float(xx[0]) / (xx[0] + xx[1])
                    mats[ii[0]].append(str(tmpf))
                    if sum(xx) < 2:
                        continue
                    cells[hh[xi]][0].append(float(xx[0]))
                    cells[hh[xi]][1].append(xx[1])
                    genes[ii[0]][0].append(float(xx[0]))
                    genes[ii[0]][1].append(xx[1])

        for ci in cells:
            if len(cells[ci][0]) < mincell:
                continue
            cfloat = sum(cells[ci][0]) / (sum(cells[ci][0]) + sum(cells[ci][1]))
            outcell.write(ci + "\t" + str(cfloat) + "\n")

        for gi in genes:
            if len(genes[gi][0]) < mingene:
                continue
            gfloat = sum(genes[gi][0]) / (sum(genes[gi][0]) + sum(genes[gi][1]))
            outgene.write(gi + "\t" + str(gfloat) + "\n")

        for mi in mats:
            outmat1.write(mi + "\t" + "\t".join(mats[mi]) + "\n")

        outcell.close()
        outgene.close()
        outmat1.close()

    @utils.add_log
    def replacment_plot(self, sample):
        import plotly
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots

        outpre = os.path.basename(sample)
        df1 = pd.read_table(sample + ".fraction_of_newRNA_per_gene.txt", header=None)
        df1.columns = ["gene", "per"]
        df = pd.read_table(sample + ".fraction_of_newRNA_per_cell.txt", header=None)
        df.columns = ["gene", "per"]

        fig = make_subplots(rows=1, cols=2)
        fig.add_trace(
            go.Violin(
                y=df1["per"],
                box_visible=True,
                line_color="black",
                meanline_visible=True,
                fillcolor="#1f77b4",
                opacity=0.6,
                x0=outpre,
            ),
            row=1,
            col=1,
        )
        fig.add_trace(
            go.Violin(
                y=df["per"],
                box_visible=True,
                line_color="black",
                meanline_visible=True,
                fillcolor="#ff7f0e",
                opacity=0.6,
                x0=outpre,
            ),
            row=1,
            col=2,
        )

        fig.update_layout(yaxis_zeroline=True, showlegend=False)
        fig.update_layout(plot_bgcolor="#FFFFFF")
        fig.update_xaxes(showgrid=False, linecolor="black", showline=True, ticks=None)
        fig.update_yaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            title_text="Fraction new RNA (per gene)",
            row=1,
            col=1,
            rangemode="tozero",
        )
        fig.update_yaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            title_text="Fraction new RNA (per cell)",
            row=1,
            col=2,
            rangemode="tozero",
        )

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type="div")

        return div

    def report_prepare(self, outdiv):
        self.add_data(replacement=outdiv)


@utils.add_log
def replacement(args):
    with Replacement(args) as runner:
        runner.run()


def get_opts_replacement(parser, sub_program):
    parser.add_argument(
        "--bg_cov",
        type=int,
        default=1,
        help="background snp depth filter, lower than bg_cov will be discarded. Only valid in csv format",
    )
    parser.add_argument(
        "--snp_threshold",
        type=float,
        default=0.5,
        help="snp threshold filter, greater than snp_threshold will be recognized as snp. Only valid in csv format",
    )
    if sub_program:
        parser.add_argument(
            "--bam", help="bam file from conversion step", required=True
        )
        parser.add_argument(
            "--bg", help="background snp file, csv or vcf format", required=True
        )
        # parser.add_argument('--cell_keep', type=int, default=100000, help='filter cell')
        parser.add_argument(
            "--min_cell",
            type=int,
            default=10,
            help="a gene expressed in at least cells, default 10",
        )
        parser.add_argument(
            "--min_gene",
            type=int,
            default=10,
            help="at least gene num in a cell, default 10",
        )
        parser = s_common(parser)
    return parser



import subprocess
import os

from celescope.tools import utils
from celescope.tools.step import Step, s_common


class Cellranger(Step):
    """
    ## Features
    - Single cell RNA-seq Gene Expression analysis by Cellranger.
    ## Output
    - `03.assemble/{sample}` Cellranger count results.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fqs_dir = os.path.abspath(args.fqs_dir)
        self.mem = args.mem
        self.other_param = args.other_param
        self.ref_path = args.ref_path
        self.soft_path = args.soft_path

        # out files
        self.cmd_line = f"{self.outdir}/{self.sample}_cmd_line"

    @utils.add_log
    def cellranger_count(self):
        """Cellranger count"""
        cmd = (
            f"{self.soft_path} count "
            f"--id={self.sample} "
            f"--transcriptome={self.ref_path} "
            f"--fastqs={self.fqs_dir} "
            f"--sample={self.sample} "
            f"--localcores={self.thread} "
            f"--localmem={self.mem} "
        )

        if self.other_param:
            cmd += " " + self.other_param

        self.cellranger_count.logger.info(cmd)
        with open(self.cmd_line, "w") as f:
            f.write(cmd)

        cwd = os.getcwd()
        os.chdir(self.outdir)
        subprocess.check_call(cmd, shell=True)
        # change dir back to avoid can not find '03.assemble/stat.txt' error
        os.chdir(cwd)

        cmd = f"cp {self.outdir}/{self.sample}/outs/web_summary.html {self.outdir}/.."
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.cellranger_count()


def cellranger(args):
    with Cellranger(args) as runner:
        runner.run()


def get_opts_cellranger(parser, sub_program):
    parser.add_argument("--ref_path", help="reference path for cellranger")
    parser.add_argument("--soft_path", help="soft path for cellranger")
    parser.add_argument(
        "--other_param", help="Other cellranger parameters.", default=""
    )
    parser.add_argument("--mem", help="memory(G)", default=10)
    if sub_program:
        s_common(parser)
        parser.add_argument("--fqs_dir", help="fastq dir after convert", required=True)
    return parser



import os
import subprocess

import pysam
from xopen import xopen

from celescope.tools import utils
from celescope.tools.step import Step, s_common


# Template Swithcing Oligos Sequence. [16-bp cell barcode][10/12-bp UMI][TSO].
TSO = "TTTCTTATATGGG"

# Path of Whitelist File in Cellranger Directory.
WHITELIST_10X_PATH = [
    "/lib/python/cellranger/barcodes",
    "/cellranger-cs/3.0.2/lib/python/cellranger/barcodes",
]

# Chemistry: [V2/V3 whitelist, 10X umi length].
CHEMISTRY_DICT = {
    "V2": ["737K-august-2016.txt", 10],
    "V3": ["3M-february-2018.txt.gz", 12],
}


class Convert(Step):
    """
    ##Features

    - Convert barcodes and UMI to 10X format.

    Output

    - `02.convert/barcode_correspond.txt` Recording barcodes correspondence.

    - `02.convert/{sample}_S1_L001_R1_001.fastq.gz` New R1 reads as cellranger input.

    - `02.convert/{sample}_S1_L001_R2_001.fastq.gz` New R2 reads as cellranger input.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.fq2 = args.fq2
        self.whitelist_suffix = CHEMISTRY_DICT[args.tenX_chemistry][0]
        self.UMI_10X_LEN = CHEMISTRY_DICT[args.tenX_chemistry][-1]

        if not os.path.islink(args.soft_path):
            raise Exception(
                "Error soft_path\nCorrect Format:\n/home/soft/cellranger-7.1.0/cellranger"
            )

        self.version = os.path.dirname(args.soft_path).split("/")[-1].split("-")[-1]
        if int(self.version.split(".")[0]) < 4:
            self.whitelist_10X_file = (
                os.path.dirname(args.soft_path)
                + f"{WHITELIST_10X_PATH[1]}/{self.whitelist_suffix}"
            )
        else:
            self.whitelist_10X_file = (
                os.path.dirname(args.soft_path)
                + f"{WHITELIST_10X_PATH[0]}/{self.whitelist_suffix}"
            )

        self.whitelist_10X_fh = xopen(self.whitelist_10X_file, "r")
        self.sgr_tenX = {}

        # out
        self.out_fq1_file = f"{self.outdir}/{self.sample}_S1_L001_R1_001.fastq.gz"
        self.out_fq2_file = f"{self.outdir}/{self.sample}_S1_L001_R2_001.fastq.gz"
        self.barcode_convert_json = f"{self.outdir}/barcode_convert.json"

    @utils.add_log
    def write_fq1(self):
        out_fq1 = xopen(self.out_fq1_file, "w")

        with pysam.FastxFile(self.fq2) as fq2_fh:
            for entry in fq2_fh:
                name = entry.name
                attrs = name.split("_")
                sgr_barcode, sgr_umi = attrs[0], attrs[1]
                new_seq1, new_qual1 = self.convert_seq(sgr_barcode, sgr_umi)
                out_fq1.write(f"@{name}\n{new_seq1}\n+\n{new_qual1}\n")

        out_fq1.close()

    @utils.add_log
    def gzip_fq2(self):
        cmd = f"gzip -c {self.fq2} > {self.out_fq2_file}"
        subprocess.check_call(cmd, shell=True)

    def convert_seq(self, barcode_sgr, umi_sgr):
        """
        Convert sgr barcode to 10X barcode; change length of sgr UMI to UMI_10X_LEN

        Args:
            barcode_sgr: str
            umi_sgr: str
        Returns:
            new_seq1: str
            new_qual1: str
        """

        if barcode_sgr in self.sgr_tenX:
            barcode_10X = self.sgr_tenX[barcode_sgr]
        else:
            # new barcode from whitelist
            barcode_10X = self.whitelist_10X_fh.readline().strip()
            self.sgr_tenX[barcode_sgr] = barcode_10X

        umi_len_sgr = len(umi_sgr)
        if umi_len_sgr > self.UMI_10X_LEN:
            umi_10X = umi_sgr[: self.UMI_10X_LEN]
        elif umi_len_sgr < self.UMI_10X_LEN:
            umi_10X = umi_sgr + "C" * (self.UMI_10X_LEN - umi_len_sgr)
        else:
            umi_10X = umi_sgr

        new_seq1 = barcode_10X + umi_10X + TSO
        new_qual1 = "F" * len(new_seq1)

        return new_seq1, new_qual1

    @utils.add_log
    def dump_tenX_sgr_barcode_json(self):
        tenX_sgr = {}
        for sgr, tenX in self.sgr_tenX.items():
            tenX_sgr[tenX] = sgr

        utils.dump_dict_to_json(tenX_sgr, self.barcode_convert_json)

    def run(self):
        self.write_fq1()
        self.gzip_fq2()
        self.dump_tenX_sgr_barcode_json()


def convert(args):
    step_name = "convert"
    with Convert(args, step_name) as runner:
        runner.run()


def get_opts_convert(parser, sub_program):
    parser.add_argument("--soft_path", help="soft path for cellranger", required=True)
    parser.add_argument(
        "--tenX_chemistry",
        help="10X chemistry version, V2 or V3 for scRNA, V2 for VDJ",
        choices=["V2", "V3"],
        default="V2",
    )
    if sub_program:
        s_common(parser)
        parser.add_argument("--fq2", help="R2 read file", required=True)
    return parser



import re
import subprocess

from celescope.tools import utils
from celescope.tools.mkref import Mkref
from celescope.tools.step import s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step


class Star_mixin(Step):
    """
    Mixin class for STAR
    """

    def __init__(self, args, add_prefix=None, display_title=None):
        super().__init__(args, display_title)

        self.fq = args.fq
        self.genomeDir = args.genomeDir
        self.out_unmapped = args.out_unmapped
        self.debug = args.debug
        self.outFilterMatchNmin = int(args.outFilterMatchNmin)
        self.multi_max = int(args.outFilterMultimapNmax)
        self.STAR_param = args.STAR_param
        self.consensus_fq = args.consensus_fq

        # parse
        self.genome = Mkref.parse_genomeDir(self.genomeDir)
        self.stat_prefix = "Reads"
        if self.consensus_fq:
            self.stat_prefix = "UMIs"

        # out
        self.outPrefix = f"{self.outdir}/{self.sample}_"
        if add_prefix:
            self.outPrefix += add_prefix + "_"
        self.STAR_map_log = f"{self.outPrefix}Log.final.out"
        self.unsort_STAR_bam = f"{self.outPrefix}Aligned.out.bam"
        self.STAR_bam = f"{self.outPrefix}Aligned.sortedByCoord.out.bam"

    @utils.add_log
    def STAR(self):
        cmd = [
            "STAR",
            "--runThreadN",
            str(self.thread),
            "--genomeDir",
            self.genomeDir,
            "--readFilesIn",
            self.fq,
            "--outFilterMultimapNmax",
            str(self.multi_max),
            "--outFileNamePrefix",
            self.outPrefix,
            "--outSAMtype",
            "BAM",
            "Unsorted",  # controls sort by Coordinate or not
            "--outFilterMatchNmin",
            str(self.outFilterMatchNmin),
        ]
        if self.out_unmapped:
            cmd += ["--outReadsUnmapped", "Fastx"]
        if self.fq[-3:] == ".gz":
            cmd += ["--readFilesCommand", "zcat"]
        cmd = " ".join(cmd)
        if self.STAR_param:
            cmd += " " + self.STAR_param
        self.STAR.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.STAR()
        self.get_star_metrics()
        self.sort_bam()
        self.index_bam()

    @utils.add_log
    def sort_bam(self):
        utils.sort_bam(
            self.unsort_STAR_bam,
            self.STAR_bam,
            threads=self.thread,
        )

    @utils.add_log
    def index_bam(self):
        utils.index_bam(self.STAR_bam)

    def get_star_metrics(self):
        """
        step metrics
        """

        with open(self.STAR_map_log, "r") as map_log:
            # number amd percent
            unique_reads_list = []
            multi_reads_list = []
            total_reads = 0
            for line in map_log:
                if line.strip() == "":
                    continue
                if re.search(r"Uniquely mapped reads", line):
                    unique_reads_list.append(line.strip().split()[-1])
                if re.search(r"of reads mapped to too many loci", line):
                    multi_reads_list.append(line.strip().split()[-1])
                if re.search(r"Number of input reads", line):
                    total_reads = int(line.strip().split()[-1])

        unique_reads = int(unique_reads_list[0])
        multi_reads = int(multi_reads_list[0])

        self.add_metric(
            name="Genome",
            value=self.genome["genome_name"],
        )
        self.add_metric(
            name=f"Uniquely Mapped {self.stat_prefix}",
            value=unique_reads,
            total=total_reads,
            help_info="reads that mapped uniquely to the genome",
        )
        self.add_metric(
            name=f"Multi-Mapped {self.stat_prefix}",
            value=multi_reads,
            total=total_reads,
            help_info="reads that mapped to multiple locations in the genome",
        )


def get_opts_star_mixin(parser, sub_program):
    parser.add_argument(
        "--genomeDir",
        help=HELP_DICT["genomeDir"],
    )
    parser.add_argument(
        "--outFilterMatchNmin",
        help="""Default `0`. Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=50,
    )
    parser.add_argument(
        "--out_unmapped", help="Output unmapped reads.", action="store_true"
    )
    parser.add_argument("--STAR_param", help=HELP_DICT["additional_param"], default="")
    parser.add_argument(
        "--outFilterMultimapNmax",
        help="Default `1`. How many places are allowed to match a read at most.",
        default=1,
    )
    parser.add_argument(
        "--starMem", help="Default `30`. Maximum memory that STAR can use.", default=30
    )
    if sub_program:
        parser.add_argument("--fq", help="Required. R2 fastq file.", required=True)
        parser.add_argument(
            "--consensus_fq",
            action="store_true",
            help="A indicator that the input fastq has been consensused.",
        )
        parser = s_common(parser)



import os
import pathlib

from collections import defaultdict

import pandas as pd

from celescope.rna.mkref import Mkref_rna
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.__init__ import HELP_DICT


class FeatureCounts(Step):
    """
    ## Features
    - Assigning uniquely mapped reads to genomic features with FeatureCounts.
    ## Output
    - `{sample}` Numbers of reads assigned to features (or meta-features).
    - `{sample}_summary` Stat info for the overall summrization results, including number of
    successfully assigned reads and number of reads that failed to be assigned due to
    various reasons (these reasons are included in the stat info).
    - `{sample}_Aligned.sortedByCoord.out.bam.featureCounts.bam` featureCounts output BAM,
    sorted by coordinates;BAM file contains tags as following(Software Version>=1.1.8):
        - CB cell barcode
        - UB UMI
        - GN gene name
        - GX gene id
    - `{sample}_name_sorted.bam` featureCounts output BAM, sorted by read name.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # set
        self.gtf = Mkref_rna.parse_genomeDir(self.args.genomeDir)["gtf"]
        self.featureCounts_param = args.featureCounts_param

        # gtf_type
        self.gtf_types = ["exon", "gene"]

        # stats
        self.feature_log_dict = defaultdict(dict)

    @staticmethod
    def read_log(log_file):
        """
        Args:
            log_file: featureCounts log summary file
        Returns:
            log_dict: {'Assigned': 123, ...}
        """
        # skip first line
        df = pd.read_csv(
            log_file, sep="\t", header=None, names=["name", "value"], skiprows=1
        )
        log_dict = df.set_index("name")["value"].to_dict()
        return log_dict

    @utils.add_log
    def run_featureCounts(self, outdir, gtf_type):
        cmd = (
            'featureCounts '
            '-s 1 '
            f'-a {self.gtf} '
            f'-o {outdir}/{self.out_prefix.split("/")[-1]} '  # not bam
            '-R BAM '
            f'-T {self.thread} '
            f'-t {gtf_type} '
            f'{self.args.input} '
            '2>&1 '
        )
        if self.featureCounts_param:
            cmd += " " + self.featureCounts_param
        self.debug_subprocess_call(cmd)

    def run(self):
        for gtf_type in self.gtf_types:
            outdir = f"{self.outdir}/tmp/{gtf_type}"
            pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)
            # out files
            name_sorted_bam = f"{outdir}/{self.sample}_name_sorted.bam"
            input_basename = os.path.basename(self.args.input)
            featureCounts_bam = f"{outdir}/{input_basename}.featureCounts.bam"
            log_file = f"{outdir}/{self.sample}.summary"

            self.run_featureCounts(outdir, gtf_type)
            if gtf_type == self.args.gtf_type:
                samtools_runner = utils.Samtools(
                    in_bam=featureCounts_bam,
                    out_bam=featureCounts_bam,
                    threads=self.thread,
                    debug=self.debug,
                )
                samtools_runner.add_tag(self.gtf)
                samtools_runner.temp_sam2bam(by="coord")
                samtools_runner.samtools_sort(
                    in_file=featureCounts_bam,
                    out_file=name_sorted_bam,
                    by="name",
                )
            self.feature_log_dict[gtf_type] = FeatureCounts.read_log(log_file)
        self.add_metrics()
        self.clean_tmp()

    @utils.add_log
    def add_metrics(self):
        total = sum(self.feature_log_dict["exon"].values())

        Assigned_exon = self.feature_log_dict["exon"]["Assigned"]
        Assigned_intergenic = self.feature_log_dict["gene"]["Unassigned_NoFeatures"]
        """
        https://academic.oup.com/nargab/article/2/3/lqaa073/5910008
        Approximately 15% of genes had exon counts that were greater than genebody counts (by a median value of eight counts). This was due to our conservative approach of excluding reads that overlapped features in multiple genes during the read summarization step by featureCounts using the argument allowMultiOverlap=FALSE. Under this strategy, some reads were counted towards the exon count set but not the genebody count set. This happens when a read overlaps the exon in one gene and the intron of another gene—it is counted towards exon counts but not genebody counts due to its overlap of multiple genebodies but not multiple exons.
        """
        Unassigned_ambiguity = self.feature_log_dict["exon"]["Unassigned_Ambiguity"]
        Assigned_intron = (
            total - Assigned_exon - Assigned_intergenic - Unassigned_ambiguity
        )

        self.add_metric(
            name="Feature Type",
            value=self.args.gtf_type.capitalize(),
            help_info="Specified by `--gtf_type`. For snRNA-seq, you need to add `--gtf_type gene` to include reads mapped to intronic regions. Staring from CeleScope v1.12.0, the default value of gtf_type is changed from `exon` to `gene`.",
        )
        self.add_metric(
            name="Reads Assigned To Exonic Regions",
            value=Assigned_exon,
            total=total,
            help_info="Reads that can be successfully assigned to exonic regions",
        )
        self.add_metric(
            name="Reads Assigned To Intronic Regions",
            value=Assigned_intron,
            total=total,
            help_info="Reads that can be successfully assigned to intronic regions",
        )
        self.add_metric(
            name="Reads Assigned To Intergenic Regions",
            value=Assigned_intergenic,
            total=total,
            help_info="Reads that can be successfully assigned to intergenic regions",
        )
        self.add_metric(
            name="Reads Unassigned Ambiguity",
            value=Unassigned_ambiguity,
            total=total,
            help_info="Alignments that overlap two or more features",
        )

    def clean_tmp(self):
        """
        remove tmp dir
        """
        use_gt = self.args.gtf_type
        retain_dir = f"{self.outdir}/tmp/{use_gt}"
        cmd = f"mv {retain_dir}/* {self.outdir};rm -rf {self.outdir}/tmp"
        self.debug_subprocess_call(cmd)


@utils.add_log
def featureCounts(args):
    with FeatureCounts(args) as runner:
        runner.run()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument(
        "--gtf_type",
        help="Specify feature type in GTF annotation",
        default="gene",
        choices=["exon", "gene"],
    )
    parser.add_argument("--genomeDir", help=HELP_DICT["genomeDir"])
    parser.add_argument(
        "--featureCounts_param", help=HELP_DICT["additional_param"], default=""
    )

    if sub_program:
        parser.add_argument("--input", help="Required. BAM file path.", required=True)
        parser = s_common(parser)
    return parser



from celescope.tag.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_tag(Multi):
    """
    `multi_tag` is used to analyze Single-cell Multiplexing data generated with CLindex<sup>TM</sup> Sample Multiplexing kits. 
    Before running `multi_tag`, you need to run scRNA-Seq data with CeleScope first.

    ## Usage

    ```
    multi_tag \\
        --mapfile ./tag.mapfile\\
        --barcode_fasta ./tag_barcode.fasta\\
        --fq_pattern L25C15\\
        --split_matrix\\
        --mod shell
    ```

    """

    def mapping_tag(self, sample):
        step = "mapping_tag"
        fq = (
            f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        )
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f"{cmd_line} " f"--fq {fq} "
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def count_tag(self, sample):
        step = "count_tag"
        read_count_file = (
            f'{self.outdir_dic[sample]["mapping_tag"]}/{sample}_read_count.tsv'
        )
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f"{cmd_line} "
            f"--match_dir {self.col4_dict[sample]} "
            f"--read_count_file {read_count_file} "
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def analysis_tag(self, sample):
        step = "analysis_tag"
        tsne_tag_file = f'{self.outdir_dic[sample]["count_tag"]}/{sample}_tsne_tag.tsv'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f"{cmd_line} "
            f"--match_dir {self.col4_dict[sample]} "
            f"--tsne_tag_file {tsne_tag_file} "
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def split_tag(self, sample):
        step = "split_tag"
        umi_tag_file = f'{self.outdir_dic[sample]["count_tag"]}/{sample}_umi_tag.tsv'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f"{cmd_line} "
            f"--match_dir {self.col4_dict[sample]} "
            f"--umi_tag_file {umi_tag_file} "
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)


def main():
    multi = Multi_tag(__ASSAY__)
    multi.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#

"""
Simple Good-Turing estimator.
Based on S implementation in
  William A. Gale & Geoffrey Sampson (1995) Good-turing frequency estimation without tears,
  Journal of Quantitative Linguistics, 2:3, 217-237, DOI: 10.1080/09296179508590051
"""

import numpy as np
import scipy.stats as sp_stats


class SimpleGoodTuringError(Exception):
    pass


def _averaging_transform(r, nr):
    d = np.concatenate((np.ones(1, dtype=int), np.diff(r)))
    dr = np.concatenate(
        (
            0.5 * (d[1:] + d[0:-1]),
            np.array((d[-1],), dtype=float),
        )
    )
    return nr.astype(float) / dr


def _rstest(r, coef):
    return r * np.power(1 + 1 / r, 1 + coef)


def simple_good_turing(xr, xnr):
    """Make a Simple Good-Turing estimate of the frequencies.
    Args:
      xr (np.array(int)): Non-zero item frequencies
      xnr (np.array(int)): Non-zero frequencies of frequencies
    Returns:
      (rstar (np.array(float)), p0 (float)):
        rstar: The adjusted non-zero frequencies
        p0: The total probability of unobserved items
    """

    xr = xr.astype(float)
    xnr = xnr.astype(float)

    xN = np.sum(xr * xnr)

    # Get Linear Good-Turing estimate
    xnrz = _averaging_transform(xr, xnr)
    slope, _intercept, _, _, _ = sp_stats.linregress(np.log(xr), np.log(xnrz))

    if slope > -1:
        raise SimpleGoodTuringError(
            "The log-log slope is > -1 (%d); the SGT estimator is not applicable to these data."
            % slope
        )

    xrst = _rstest(xr, slope)
    xrstrel = xrst / xr

    # Get traditional Good-Turing estimate
    xrtry = xr == np.concatenate((xr[1:] - 1, np.zeros(1)))
    xrstarel = np.zeros(len(xr))
    xrstarel[xrtry] = (
        (xr[xrtry] + 1)
        / xr[xrtry]
        * np.concatenate((xnr[1:], np.zeros(1)))[xrtry]
        / xnr[xrtry]
    )

    # Determine when to switch from GT to LGT estimates
    tursd = np.ones(len(xr))
    for i in range(len(xr)):
        if xrtry[i]:
            tursd[i] = (
                float(i + 2) / xnr[i] * np.sqrt(xnr[i + 1] * (1 + xnr[i + 1] / xnr[i]))
            )

    xrstcmbrel = np.zeros(len(xr))
    useturing = True
    for r in range(len(xr)):
        if not useturing:
            xrstcmbrel[r] = xrstrel[r]
        else:
            if np.abs(xrstrel[r] - xrstarel[r]) * (1 + r) / tursd[r] > 1.65:
                xrstcmbrel[r] = xrstarel[r]
            else:
                useturing = False
                xrstcmbrel[r] = xrstrel[r]

    # Renormalize the probabilities for observed objects
    sumpraw = np.sum(xrstcmbrel * xr * xnr / xN)

    xrstcmbrel = xrstcmbrel * (1 - xnr[0] / xN) / sumpraw
    p0 = xnr[0] / xN

    return (xr * xrstcmbrel, p0)


def sgt_proportions(frequencies):
    """Use Simple Good-Turing estimate to adjust for unobserved items
    Args:
      frequencies (np.array(int)): Nonzero frequencies of items
    Returns:
      (pstar (np.array(float)), p0 (float)):
        pstar: The adjusted non-zero proportions
        p0: The total probability of unobserved items
    """
    if len(frequencies) == 0:
        raise ValueError("Input frequency vector is empty")
    if np.count_nonzero(frequencies) != len(frequencies):
        raise ValueError("Frequencies must be greater than zero")

    freqfreqs = np.bincount(frequencies)
    assert freqfreqs[0] == 0
    use_freqs = np.flatnonzero(freqfreqs)

    if len(use_freqs) < 10:
        raise SimpleGoodTuringError(
            "Too few non-zero frequency items (%d). Aborting SGT." % len(use_freqs)
        )

    rstar, p0 = simple_good_turing(use_freqs, freqfreqs[use_freqs])

    # rstar contains the smoothed frequencies.
    # Map each original frequency r to its smoothed rstar.
    rstar_dict = dict(zip(use_freqs, rstar))

    rstar_sum = np.sum(freqfreqs[use_freqs] * rstar)
    rstar_i = np.fromiter(
        (rstar_dict[f] for f in frequencies), dtype=float, count=len(frequencies)
    )
    pstar = (1 - p0) * (rstar_i / rstar_sum)

    assert np.isclose(p0 + np.sum(pstar), 1)
    return (pstar, p0)


def test_prosody():
    data = (
        (1, 120),
        (2, 40),
        (3, 24),
        (4, 13),
        (5, 15),
        (6, 5),
        (7, 11),
        (8, 2),
        (9, 2),
        (10, 1),
        (12, 3),
        (14, 2),
        (15, 1),
        (16, 1),
        (17, 3),
        (19, 1),
        (20, 3),
        (21, 2),
        (23, 3),
        (24, 3),
        (25, 3),
        (26, 2),
        (27, 2),
        (28, 1),
        (31, 2),
        (32, 2),
        (33, 1),
        (34, 2),
        (36, 2),
        (41, 3),
        (43, 1),
        (45, 3),
        (46, 1),
        (47, 1),
        (50, 1),
        (71, 1),
        (84, 1),
        (101, 1),
        (105, 1),
        (121, 1),
        (124, 1),
        (146, 1),
        (162, 1),
        (193, 1),
        (199, 1),
        (224, 1),
        (226, 1),
        (254, 1),
        (257, 1),
        (339, 1),
        (421, 1),
        (456, 1),
        (481, 1),
        (483, 1),
        (1140, 1),
        (1256, 1),
        (1322, 1),
        (1530, 1),
        (2131, 1),
        (2395, 1),
        (6925, 1),
        (7846, 1),
    )

    # Computed using R 3.5.1 w/ the Gale S code
    expect_p0 = 0.003883244
    expect_rstar = np.array(
        (
            0.7628079,
            1.706448,
            2.679796,
            3.663988,
            4.653366,
            5.645628,
            6.63966,
            7.634856,
            8.63086,
            9.627446,
            11.62182,
            13.61725,
            14.61524,
            15.61336,
            16.6116,
            18.60836,
            19.60685,
            20.6054,
            22.60264,
            23.60133,
            24.60005,
            25.5988,
            26.59759,
            27.59639,
            30.59294,
            31.59183,
            32.59073,
            33.58964,
            35.58751,
            40.58235,
            42.58035,
            44.57836,
            45.57738,
            46.57641,
            49.57351,
            70.55399,
            83.54229,
            100.5272,
            104.5237,
            120.5097,
            123.507,
            145.4879,
            161.474,
            192.4472,
            198.4421,
            223.4205,
            225.4188,
            253.3947,
            256.3922,
            338.3218,
            420.2514,
            455.2215,
            480.2,
            482.1983,
            1138.636,
            1254.537,
            1320.48,
            1528.302,
            2128.788,
            2392.562,
            6918.687,
            7838.899,
        )
    )

    xr = np.array([d[0] for d in data], dtype=int)
    xnr = np.array([d[1] for d in data], dtype=int)

    rstar, p0 = simple_good_turing(xr, xnr)

    assert np.abs(p0 - expect_p0) < 1e-9
    assert np.all(np.abs(rstar - expect_rstar) < 1e-3)
    assert np.all((np.abs(rstar - expect_rstar)) / expect_rstar < 1e-4)



import pandas as pd
import subprocess
import glob

from celescope.flv_trust4.summarize import Summarize
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.flv_trust4.__init__ import TOOLS_DIR
from celescope.tools.plotly_plot import Bar_plot


CELL_TYPE_DICT = {
    "TCR": ["T_cells", "NKT_cells", "T cells", "NK T cells", "Tcells"],
    "BCR": [
        "Plasma_cells",
        "B_cells",
        "Mature_B_cell",
        "Plasma cells",
        "B cells",
        "Bcells",
    ],
}


class Annotation(Step):
    """
    ## Features

    - Output assembled T/B cells mapping to transcriptome if rds and auto-assign info exist in match directory.
    - Generate VDJ annotation info, clonetypes table and bar-plot of clonetypes distribution in html.

    ## Output
    - `05.annotation/{sample}_assign.png` Umap plot of Auto-assigned celltype in transcriptome.
    - `05.annotation/{sample}_cluster_umap.png` Umap plot of Cluster in transcriptome.
    - `05.annotation/{sample}_umapplot.png` Umap plot of assembled barcodes marked as read color.
    - `05.annotation/{sample}_distribution.txt` Number of assembled barcodes in every clusters.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.match_dir = args.match_dir
        self.chains, self.paired_groups = Summarize._parse_seqtype(self.seqtype)
        self.contig_file = f"{args.summarize_out}/{self.sample}_filtered_contig.csv"
        self.clonotype_file = f"{args.summarize_out}/clonotypes.csv"
        self.celltype_set = CELL_TYPE_DICT[self.seqtype]

        try:
            self.rds = glob.glob(f"{self.match_dir}/06.analysis/*.rds")[0]
            self.assign_file = glob.glob(
                f"{self.match_dir}/06.analysis/*_auto_assign/*_auto_cluster_type.tsv"
            )[0]
        except IndexError:
            pass

    def parse_clonotype(self):
        """Generate clonotypes table in html."""
        df_clonotypes = pd.read_csv(self.clonotype_file, sep=",")
        df_clonotypes["ClonotypeID"] = df_clonotypes["clonotype_id"].apply(
            lambda x: x.strip("clonetype")
        )
        df_clonotypes["Frequency"] = df_clonotypes["frequency"]
        df_clonotypes["Proportion"] = df_clonotypes["proportion"].apply(
            lambda x: f"{round(x*100, 2)}%"
        )
        df_clonotypes["CDR3_aa"] = df_clonotypes["cdr3s_aa"].apply(
            lambda x: x.replace(";", "<br>")
        )

        return df_clonotypes

    @staticmethod
    @utils.add_log
    def get_vdj_metric(df, chains, pairs):
        """
        Add vdj metrics in html.
        """
        metric_result = []
        fl_pro_pair_df = pd.DataFrame(
            df[df["productive"] == True].barcode.value_counts()
        )
        fl_pro_pair_df = fl_pro_pair_df[fl_pro_pair_df["barcode"] >= 2]
        cell_nums = len(set(df["barcode"]))

        metric_result.append(
            {
                "name": "Cells With Productive V-J Spanning Pair",
                "value": fl_pro_pair_df.shape[0],
                "total": cell_nums,
            }
        )

        for pair in pairs:
            chain1, chain2 = pair.split("_")[0], pair.split("_")[1]
            cbs1 = set(
                df[
                    (df["full_length"] == True)
                    & (df["productive"] == True)
                    & (df["chain"] == chain1)
                ].barcode
            )
            cbs2 = set(
                df[
                    (df["full_length"] == True)
                    & (df["productive"] == True)
                    & (df["chain"] == chain2)
                ].barcode
            )
            paired_cbs = len(cbs1.intersection(cbs2))

            metric_result.append(
                {
                    "name": f"Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair",
                    "value": paired_cbs,
                    "total": cell_nums,
                    "help_info": "Fraction of cell-associated barcodes with one productive contig for each chain of the receptor pair.A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region",
                }
            )

        for chain in chains:
            metric_result.append(
                {
                    "name": f"Cells With {chain} Contig",
                    "value": len(set(df[df["chain"] == chain].barcode)),
                    "total": cell_nums,
                    "help_info": f"Fraction of cell-associated barcodes with at least one {chain} contig annotated as a full or partial V(D)J gene",
                }
            )
            metric_result.append(
                {
                    "name": f"Cells With CDR3-annotated {chain} Contig",
                    "value": len(
                        set(df[(df["chain"] == chain) & (df["cdr3"] != None)].barcode)
                    ),
                    "total": cell_nums,
                }
            )
            metric_result.append(
                {
                    "name": f"Cells With V-J Spanning {chain} Contig",
                    "value": len(
                        set(
                            df[
                                (df["full_length"] == True) & (df["chain"] == chain)
                            ].barcode
                        )
                    ),
                    "total": cell_nums,
                    "help_info": f"Fraction of cell-associated barcodes with at least one contig spanning the 5' end of the V region to the 3' end of the J region for {chain}",
                }
            )
            metric_result.append(
                {
                    "name": f"Cells With Productive {chain} Contig",
                    "value": len(
                        set(
                            df[
                                (df["full_length"] == True)
                                & (df["productive"] == True)
                                & (df["chain"] == chain)
                            ].barcode
                        )
                    ),
                    "total": cell_nums,
                    "help_info": "Fraction of cell-associated barcodes with productive IGL chain. A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region",
                }
            )

        return metric_result

    @utils.add_log
    def annotation_process(self):
        """Add metrics, clonotypes table, and bar-plot of clonotypes distribution in html."""
        df = pd.read_csv(self.contig_file)

        annotation_summary = Annotation.get_vdj_metric(
            df, self.chains, self.paired_groups
        )
        for anno in annotation_summary:
            self.add_metric(
                anno["name"], anno["value"], anno.get("total"), anno.get("help_info")
            )

        df_clonotypes = self.parse_clonotype()
        title = "Clonetypes"
        table_dict = self.get_table_dict(
            title=title,
            table_id="clonetypes",
            df_table=df_clonotypes[
                ["ClonotypeID", "CDR3_aa", "Frequency", "Proportion"]
            ],
        )
        self.add_data(table_dict=table_dict)

        df_clonotypes["ClonotypeID"] = df_clonotypes["ClonotypeID"].astype("int")
        df_clonotypes.sort_values(by=["ClonotypeID"], inplace=True)
        Barplot = Bar_plot(df_bar=df_clonotypes).get_plotly_div()
        self.add_data(Barplot=Barplot)

    @staticmethod
    @utils.add_log
    def run_mapping(rds, contig, sample, outdir, assign):
        cmd = (
            f"Rscript {TOOLS_DIR}/VDJmapping.R "
            f"--rds {rds} "
            f"--VDJ {contig} "
            f"--sample {sample} "
            f"--outdir {outdir} "
            f"--assign_file {assign} "
            "2>&1 "
        )
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def mapping_process(self):
        """Mapping result with matched scRNA."""
        Annotation.run_mapping(
            self.rds, self.contig_file, self.sample, self.outdir, self.assign_file
        )

        meta = pd.read_csv(f"{self.outdir}/{self.sample}_meta.csv")
        metaTB = meta[meta["CellTypes"].isin(self.celltype_set)]
        mappedmeta = meta[meta["Class"] == "T/BCR"]
        mappedmetaTB = mappedmeta[mappedmeta["CellTypes"].isin(self.celltype_set)]

        self.add_metric(
            "Total Cell Number in Matched transcriptome", meta.shape[0], show=False
        )
        self.add_metric(
            "Cell Number Successfully Mapped to transcriptome",
            mappedmeta.shape[0],
            show=False,
        )
        self.add_metric(
            "T/B cell Number in Matched transcriptome", metaTB.shape[0], show=False
        )
        self.add_metric(
            "Cell Number Successfully Mapped to T/B cell in transcriptome",
            mappedmetaTB.shape[0],
            show=False,
        )

    def run(self):
        self.annotation_process()

        try:
            if self.rds and self.assign_file:
                self.mapping_process()
        except AttributeError:
            print("rds file and type file do not exist" + "\n")
        except ZeroDivisionError:
            print("Not found auto-assigned T/B cell in matched sc-RNA")


@utils.add_log
def annotation(args):
    with Annotation(args, display_title="V(D)J Annotation") as runner:
        runner.run()


def get_opts_annotation(parser, sub_program):
    parser.add_argument(
        "--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True
    )
    if sub_program:
        s_common(parser)
        parser.add_argument(
            "--match_dir", help="scRNA-seq match directory", required=True
        )
        parser.add_argument(
            "--summarize_out", help="summarize output directory", required=True
        )
    return parser



__STEPS__ = [
    "sample",
    "barcode",
    "cutadapt",
    "star",
    "featureCounts",
    "count",
    "analysis",
    "conversion",
    "substitution",
    "replacement",
    "replace_tsne",
]

__ASSAY__ = "dynaseq"

IMPORT_DICT = {
    "star": "celescope.rna",
    "analysis": "celescope.rna",
}

# m: memory
# x: thread
RESOURCE = {
    "sample": {"m": 1, "x": 1},
    "barcode": {"m": 5, "x": 1},
    "cutadapt": {"m": 5, "x": 1},
    "star": {"m": 30, "x": 1},
}



# barcode
PATTERN_DICT = {
    "auto": None,
    "scopeV1": "C12U8T18",
    "scopeV2.0.0": "C8L16C8L16C8U8T18",
    "scopeV2.0.1": "C8L16C8L16C8L1U8T18",
    "scopeV2.1.0": "C8L16C8L16C8U12T18",
    "scopeV2.1.1": "C8L16C8L16C8L1U12T18",
    "scopeV2.2.1": "C8L16C8L16C8L1U12T18",
    "scopeV3.0.1": "C9L16C9L16C9L1U12T18",
    "GEXSCOPE-V3": "C9L6C9L6C9L1U12",
    "flv_rna": "C8L16C8L16C8U9",
    "flv": "U9C8L16C8L16C8",
    "customized": None,
}


# count
RAW_MATRIX_DIR_SUFFIX = ["raw_feature_bc_matrix", "all_matrix"]
FILTERED_MATRIX_DIR_SUFFIX = ["filtered_feature_bc_matrix", "matrix_10X"]
MATRIX_FILE_NAME = "matrix.mtx"
FEATURE_FILE_NAME = "genes.tsv"
BARCODE_FILE_NAME = "barcodes.tsv"

# mkref
GENOME_CONFIG = "celescope_genome.config"



#!/usr/bin/env python
# v1.0

import pysam
import os
import subprocess
import numpy as np
import pandas as pd
from collections import Counter
from celescope.tools.step import Step, s_common
from celescope.tools import utils


class Conversion(Step):
    """
    ## Features
    - Get conversion pos in each read.
        - Get snp info.

    ## Output
    - `{sample}.PosTag.bam` Bam file with conversion info.
    - `{sample}.PosTag.csv` SNP info in csv format.
    """

    def __init__(self, args):
        Step.__init__(self, args)
        # input files
        # self.ifile = os.path.join(args.outdir, args.sample+'.bam')
        self.sample = args.sample
        self.strandednessfile = args.strand
        self.inbam = args.bam
        self.bcfile = args.cell
        self.outdir = args.outdir
        self.thread = args.thread
        self.qual = int(args.basequalilty)

        # output files
        self.outfile_bam = os.path.join(args.outdir, args.sample + ".PosTag.bam")
        self.outfile_csv = os.path.join(args.outdir, args.sample + ".PosTag.csv")

    @utils.add_log
    def run(self):
        # Adding tags
        ContigLocs, AnnoteLocs = self.addTags(
            self.inbam,
            self.outfile_bam,
            self.bcfile,
            self.strandednessfile,
            self.qual,
            self.thread,
        )
        cmd = ["samtools index", self.outfile_bam]
        self.run_cmd(cmd)

        # Obtaining conversion positions
        bam = pysam.AlignmentFile(self.outfile_bam, "rb")
        # ContigLocs, AnnoteLocs = self.CountConvperPos(bam)

        # Obtaining coverage over conversion position
        ConvsPerPos, CoverofPosWithConvs = self.CountReadConverPerConvPos(
            bam, ContigLocs
        )
        A = self.ExportasVcf(ConvsPerPos, CoverofPosWithConvs, AnnoteLocs)
        A["sample"] = self.sample
        # Saving result
        A.to_csv(self.outfile_csv)
        bam.close()

        # cmd = ['rm', self.ifile]
        # self.run_cmd(cmd)
        # cmd = ['rm', self.ifile+'.bai']
        # self.run_cmd(cmd)

    def run_cmd(self, cmd):
        subprocess.call(" ".join(cmd), shell=True)

    # @utils.add_log
    def CountConvperPos_read(self, read, ContigLocs, AnnoteLocs):
        # ContigLocs = {}
        # AnnoteLocs = {}
        try:
            if read.get_tag("ST") == "+":
                locs = read.get_tag("TL")
            else:
                locs = read.get_tag("AL")
            if locs[0] != 0:
                if read.reference_name in ContigLocs:
                    ContigLocs[read.reference_name].extend(locs)
                else:
                    ContigLocs[read.reference_name] = list(locs)
                if read.reference_name not in AnnoteLocs:
                    for i, each in enumerate(locs):
                        if i == 0:
                            AnnoteLocs[read.reference_name] = {each: read.get_tag("GX")}
                        else:
                            AnnoteLocs[read.reference_name][each] = read.get_tag("GX")
                else:
                    for i, each in enumerate(locs):
                        if each not in AnnoteLocs[read.reference_name]:
                            AnnoteLocs[read.reference_name][each] = read.get_tag("GX")
        except (ValueError, KeyError):
            print("Reads Error {}".format(str(read.qurey_name)))
        return ContigLocs, AnnoteLocs

    @utils.add_log
    def CountReadConverPerConvPos(self, bam, ContigLocs):
        ConvsPerPos = {}
        CoverofPosWithConvs = {}
        for key in ContigLocs.keys():
            ConvsPerPos[key] = {}
            ContigLocs[key] = sorted(ContigLocs[key])
            ConvsPerPos[key] = Counter(ContigLocs[key])

            CoverofPosWithConvs[key] = {}
            for key2 in ConvsPerPos[key].keys():
                try:
                    CoverofPosWithConvs[key][key2] = bam.count(key, key2, key2 + 1)
                except ValueError:
                    continue
        return ConvsPerPos, CoverofPosWithConvs

    @utils.add_log
    def ExportasVcf(self, ConvsPerPos, CoverofPosWithConvs, AnnoteLocs):
        # Chrom, Pos , ConvsPerPs, CoverofPosWithConvs
        order_columns = ["chrom", "pos", "convs", "covers", "posratio", "gene_id"]
        Outputdf = pd.DataFrame(columns=order_columns)
        for key in ConvsPerPos.keys():
            if len(ConvsPerPos[key]) == 0:
                continue
            df = pd.DataFrame.from_dict(ConvsPerPos[key], orient="index")
            df1 = pd.DataFrame.from_dict(CoverofPosWithConvs[key], orient="index")
            df3 = pd.DataFrame.from_dict(AnnoteLocs[key], orient="index")
            # df.index.name = 'pos'
            # df1.index.name = 'pos'
            df.columns = ["convs"]
            df1.columns = ["covers"]
            df3.columns = ["gene_id"]
            df2 = pd.concat([df, df1, df3], axis=1)
            df2["pos"] = df2.index
            # df2.index = np.arange(df2.shape[0])
            df2["chrom"] = np.repeat(key, df2.shape[0])
            df2["posratio"] = df2["convs"] / df2["covers"]
            Outputdf = pd.concat([Outputdf, df2[order_columns]])
        return Outputdf.reset_index(drop=True)

    def createTag(self, d):
        return "".join(["".join(key) + str(d[key]) + ";" for key in d.keys()])[:-1]

    def convInRead(self, read, qual=20):
        specific_conversions = {}
        total_content = {"a": 0, "c": 0, "g": 0, "t": 0}
        specific_conversions[("c", "A")] = 0
        specific_conversions[("g", "A")] = 0
        specific_conversions[("t", "A")] = 0
        specific_conversions[("a", "C")] = 0
        specific_conversions[("g", "C")] = 0
        specific_conversions[("t", "C")] = 0
        specific_conversions[("a", "G")] = 0
        specific_conversions[("c", "G")] = 0
        specific_conversions[("t", "G")] = 0
        specific_conversions[("a", "T")] = 0
        specific_conversions[("c", "T")] = 0
        specific_conversions[("g", "T")] = 0
        specific_conversions[("a", "N")] = 0
        specific_conversions[("c", "N")] = 0
        specific_conversions[("g", "N")] = 0
        specific_conversions[("t", "N")] = 0

        tC_loc = []
        aG_loc = []

        try:
            refseq = read.get_reference_sequence().lower()
        except UnicodeDecodeError:
            refseq = ""
        except AssertionError:
            return 0

        for base in total_content.keys():
            total_content[base] += refseq.count(base)
        for pair in read.get_aligned_pairs(with_seq=True):
            try:
                if pair[0] is not None and pair[1] is not None and pair[2] is not None:
                    if (
                        str(pair[2]).islower()
                        and not read.query_qualities[pair[0]] < qual
                    ):
                        specific_conversions[(pair[2], read.seq[pair[0]])] += 1
                        if (pair[2], read.seq[pair[0]]) == ("t", "C"):
                            tC_loc.append(pair[1])
                        if (pair[2], read.seq[pair[0]]) == ("a", "G"):
                            aG_loc.append(pair[1])
            except (UnicodeDecodeError, KeyError):
                continue
        SC_tag = self.createTag(specific_conversions)
        TC_tag = self.createTag(total_content)

        if len(tC_loc) == 0:
            tC_loc.append(0)
        if len(aG_loc) == 0:
            aG_loc.append(0)
        return SC_tag, TC_tag, tC_loc, aG_loc

    @utils.add_log
    def addTags(
        self, bamfilename, outfile_bam, cellfile, strandednessfile, qual=20, thread=8
    ):
        ContigLocs = {}
        AnnoteLocs = {}
        bamfile = pysam.AlignmentFile(bamfilename, "rb")
        mod_bamfile = pysam.AlignmentFile(outfile_bam, mode="wb", template=bamfile)
        tmpread = {}
        cellist = pd.read_csv(cellfile, header=None, index_col=0)
        cells = cellist.index
        strandedness = pd.read_csv(strandednessfile, header=None, index_col=0)
        genes = strandedness.index
        chrp = ""

        class GeneError(Exception):
            pass

        for read in bamfile.fetch(until_eof=True):
            try:
                if not read.has_tag("GX"):
                    continue
                if read.get_tag("CB") not in cells:
                    continue
                if read.get_tag("GX") not in genes:
                    raise GeneError
                readid = read.get_tag("CB") + read.get_tag("UB") + read.get_tag("GX")
                chr1 = read.reference_name

                if chr1 != chrp and chrp != "":
                    for x in tmpread:
                        mod_bamfile.write(tmpread[x][1])
                        ContigLocs, AnnoteLocs = self.CountConvperPos_read(
                            tmpread[x][1], ContigLocs, AnnoteLocs
                        )
                    tmpread = {}

                tags = self.convInRead(read, qual)
                if tags == 0:
                    continue
                read.set_tag("SC", tags[0], "Z")
                read.set_tag("TC", tags[1], "Z")
                read.set_tag("TL", tags[2])
                read.set_tag("AL", tags[3])
                read.set_tag("ST", strandedness.loc[read.get_tag("XT")][1])

                if read.get_tag("ST") == "+":
                    stag = read.get_tag("TL")
                else:
                    stag = read.get_tag("AL")
                if len(stag) == 1 and stag[0] == 0:
                    tc = 0
                else:
                    tc = len(stag)
                if readid in tmpread:
                    if tc > tmpread[readid][0]:
                        tmpread[readid] = [tc, read]
                else:
                    tmpread[readid] = [tc, read]

            except (ValueError, KeyError):
                continue
            except GeneError:
                print(
                    "{} is not in strand file, please check your files.".format(
                        read.get_tag("GX")
                    )
                )
                continue
            # except (Exception):
            #    print("convert error")
            #    sys.exit(1)

        for x in tmpread:
            mod_bamfile.write(tmpread[x][1])
            ContigLocs, AnnoteLocs = self.CountConvperPos_read(
                tmpread[x][1], ContigLocs, AnnoteLocs
            )

        bamfile.close()
        mod_bamfile.close()

        cmd = ["samtools sort -@", str(thread), "-o", outfile_bam + ".bam", outfile_bam]
        self.run_cmd(cmd)
        cmd = ["mv", outfile_bam + ".bam", outfile_bam]
        self.run_cmd(cmd)
        return ContigLocs, AnnoteLocs


@utils.add_log
def conversion(args):
    with Conversion(args) as runner:
        runner.run()


def get_opts_conversion(parser, sub_program):
    parser.add_argument(
        "--strand", help='gene strand file, the format is "geneID,+/-"', required=True
    )
    parser.add_argument(
        "--basequalilty",
        default=20,
        type=int,
        help="min base quality of the read sequence",
        required=False,
    )
    if sub_program:
        parser.add_argument(
            "--bam",
            help='featureCount bam(sortedByCoord), must have "MD" tag, set in star step',
            required=True,
        )
        parser.add_argument("--cell", help="barcode cell list", required=True)
        parser = s_common(parser)
    return parser



#!/bin/env python
# coding=utf8

import os
import pysam
import re
import pandas as pd
import plotly
import plotly.graph_objects as go
from celescope.tools.step import Step, s_common
from celescope.tools import utils


class Substitution(Step):
    """
    ## Features
    - Computes the overall conversion rates in reads and plots a barplot.

    ## Output
    - `{sample}.substitution.txt` Tab-separated table of the overall conversion rates.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # input files
        self.sample = args.sample
        self.bam_file = args.bam
        self.outdir = args.outdir

        # output files
        self.outstat = os.path.join(self.outdir, self.sample + ".substitution.txt")

    @utils.add_log
    def run(self):
        # overall rate
        for_base, rev_base, is_forward, is_reverse = self.get_sub_tag(self.bam_file)
        self.sub_stat(for_base, rev_base, is_forward, is_reverse, self.outstat)
        div_item = self.sub_plot(self.outstat)

        self.report_prepare(div_item)

    @utils.add_log
    def get_sub_tag(self, bam):
        bamfile = pysam.AlignmentFile(bam, "rb")
        is_reverse = {
            "cA": 0,
            "gA": 0,
            "tA": 0,
            "aC": 0,
            "gC": 0,
            "tC": 0,
            "aG": 0,
            "cG": 0,
            "tG": 0,
            "aT": 0,
            "cT": 0,
            "gT": 0,
        }
        is_forward = {
            "cA": 0,
            "gA": 0,
            "tA": 0,
            "aC": 0,
            "gC": 0,
            "tC": 0,
            "aG": 0,
            "cG": 0,
            "tG": 0,
            "aT": 0,
            "cT": 0,
            "gT": 0,
        }
        for_base = {"a": 0, "c": 0, "g": 0, "t": 0}
        rev_base = {"a": 0, "c": 0, "g": 0, "t": 0}
        snp_tags = [
            "",
            "cA",
            "gA",
            "tA",
            "aC",
            "gC",
            "tC",
            "aG",
            "cG",
            "tG",
            "aT",
            "cT",
            "gT",
        ]
        ref_tags = ["", "a", "c", "g", "t"]
        for read in bamfile.fetch():
            try:
                snpmatch = re.match(
                    r"cA(\d+);gA(\d+);tA(\d+);aC(\d+);gC(\d+);tC(\d+);aG(\d+);cG(\d+);tG(\d+);aT(\d+);cT(\d+);gT(\d+);",
                    read.get_tag("SC"),
                    re.M,
                )
                totmatch = re.match(
                    r"a(\d+);c(\d+);g(\d+);t(\d+)", read.get_tag("TC"), re.M
                )
                if snpmatch and totmatch:
                    if read.is_reverse:
                        for j in range(1, len(ref_tags)):
                            rev_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1, len(snp_tags)):
                            is_reverse[snp_tags[i]] += int(snpmatch.group(i))
                    else:
                        for j in range(1, len(ref_tags)):
                            for_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1, len(snp_tags)):
                            is_forward[snp_tags[i]] += int(snpmatch.group(i))
            except (ValueError, KeyError):
                continue
        bamfile.close()

        return for_base, rev_base, is_forward, is_reverse

    @utils.add_log
    def sub_stat(self, for_base, rev_base, is_forward, is_reverse, outfile):
        convertdict = {
            "a": ["aC", "aG", "aT"],
            "c": ["cA", "cG", "cT"],
            "g": ["gA", "gC", "gT"],
            "t": ["tA", "tC", "tG"],
        }
        subdict = {
            "a": "t",
            "t": "a",
            "c": "g",
            "g": "c",
            "aC": "tG",
            "aG": "tC",
            "aT": "tA",
            "cA": "gT",
            "cG": "gC",
            "cT": "gA",
            "gA": "cT",
            "gC": "cG",
            "gT": "cA",
            "tA": "aT",
            "tC": "aG",
            "tG": "aC",
        }
        outdict = {
            "aC": "A_to_C",
            "aG": "A_to_G",
            "aT": "A_to_T",
            "cA": "C_to_A",
            "cG": "C_to_G",
            "cT": "C_to_T",
            "gA": "G_to_A",
            "gC": "G_to_C",
            "gT": "G_to_T",
            "tA": "T_to_A",
            "tC": "T_to_C",
            "tG": "T_to_G",
        }
        outw = open(outfile, "w")
        for x in ["a", "c", "g", "t"]:
            fbase = for_base[x]
            rbase = rev_base[subdict[x]]
            for y in convertdict[x]:
                fcov = is_forward[y] * 100 / float(fbase)
                rcov = is_reverse[subdict[y]] * 100 / float(rbase)
                outw.write(
                    outdict[y] + "\t" + "%.3f" % fcov + "\t" + "%.3f" % rcov + "\n"
                )
        outw.close()

    @utils.add_log
    def sub_plot(self, txt):
        df = pd.read_table(txt, header=None)
        df.columns = ["sample", "+", "-"]

        fig = go.Figure()
        # 设置颜色：
        import plotly.express as px

        num4colors = 0
        num4rainbow = 0
        colors_list = []
        while num4colors < 100:
            if num4rainbow == 9:
                num4rainbow = 0
            colors_list.append(px.colors.qualitative.Plotly[num4rainbow])
            num4colors += 1
            num4rainbow += 1

        num4sample = 0
        colors4sample = {}
        num4x = 0

        for sample in df["sample"].unique():
            legend_show = True
            colors4sample[sample] = colors_list[num4sample]
            num4sample += 1
            flag_x = "x" + str(num4x + 1)
            df_plot = df[df["sample"] == sample]
            num4x += 1

            fig.add_trace(
                go.Bar(
                    name=sample + "+",
                    x=df_plot["sample"],
                    y=df_plot["+"],
                    legendgroup=sample,
                    marker_color=colors4sample[sample],
                    marker_line_color="#FFFFFF",
                    showlegend=legend_show,
                    xaxis=flag_x,
                )
            )
            fig.add_trace(
                go.Bar(
                    name=sample + "-",
                    x=df_plot["sample"],
                    y=df_plot["-"],
                    legendgroup=sample,
                    showlegend=legend_show,
                    marker_color=colors4sample[sample],
                    marker_line_color="#FFFFFF",
                    opacity=0.3,
                    xaxis=flag_x,
                )
            )

        fig.update_layout(barmode="stack")

        per = 1 / (num4x + 1)
        gap4bar = per / len(df["sample"].unique())
        num4x = 0
        for typeB in df["sample"].unique():
            if num4x == 0:
                flag_x = "xaxis"
            else:
                flag_x = "xaxis" + str(num4x + 1)
            anchor_x = "x" + str(num4x + 1)
            num4x += 1
            fig["layout"][flag_x] = dict(
                domain=[per * num4x, per * (num4x + 1) - gap4bar],
                anchor=anchor_x,
                title=typeB,
            )

        fig.update_layout(plot_bgcolor="#FFFFFF")
        fig.update_xaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            showticklabels=False,
        )
        fig.update_yaxes(
            showgrid=False, linecolor="black", showline=True, ticks="outside"
        )
        width_num = (
            400 * (len(df["sample"].unique()) * len(df["sample"].unique())) / (5 * 12)
        )  # 控制柱形图的宽度
        fig.update_layout(height=500, width=width_num)
        fig.update_layout(legend=dict(orientation="h"))
        fig.update_layout(
            legend=dict(
                yanchor="top",
                y=1.3,
                xanchor="left",
                x=0.05,
                valign="top",
            )
        )

        fig.update_layout(
            yaxis_title="Rates of nucleotide substitution (%)",
        )
        fig.update_xaxes(tickangle=-80, title_font={"size": 15}, title_standoff=25)

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type="div")

        return div

    def report_prepare(self, outdiv):
        self.add_data(substitution=outdiv)


@utils.add_log
def substitution(args):
    with Substitution(args) as runner:
        runner.run()


def get_opts_substitution(parser, sub_program):
    if sub_program:
        parser.add_argument(
            "--bam", help="bam file from conversion step", required=True
        )
        parser = s_common(parser)
    return parser



from collections import defaultdict
from itertools import groupby

import pandas as pd
import pysam

from celescope.tools import utils
from celescope.tools.count import Count, get_opts_count


class Count_capture_rna(Count):
    def bam2table(self):
        """
        read probe file
        """
        probe_gene_count_dict = utils.genDict(dim=4, valType=int)

        samfile = pysam.AlignmentFile(self.bam, "rb")
        with open(self.count_detail_file, "wt") as fh1:
            fh1.write("\t".join(["Barcode", "geneID", "UMI", "count"]) + "\n")

            def keyfunc(x):
                return x.query_name.split("_", 1)[0]

            for _, g in groupby(samfile, keyfunc):
                gene_umi_dict = defaultdict(lambda: defaultdict(int))
                for seg in g:
                    (barcode, umi, probe) = seg.query_name.split("_")[:3]
                    if probe != "None":
                        probe_gene_count_dict[probe]["total"][barcode][umi] += 1
                        if seg.has_tag("XT"):
                            geneID = seg.get_tag("XT")
                            geneName = self.gtf_dict[geneID]
                            probe_gene_count_dict[probe][geneName][barcode][umi] += 1
                        else:
                            probe_gene_count_dict[probe]["None"][barcode][umi] += 1
                    if not seg.has_tag("XT"):
                        continue
                    geneID = seg.get_tag("XT")
                    gene_umi_dict[geneID][umi] += 1
                for gene_id in gene_umi_dict:
                    Count.correct_umi(gene_umi_dict[gene_id])

                # output
                for gene_id in gene_umi_dict:
                    for umi in gene_umi_dict[gene_id]:
                        fh1.write(
                            "%s\t%s\t%s\t%s\n"
                            % (barcode, gene_id, umi, gene_umi_dict[gene_id][umi])
                        )

        # out probe
        row_list = []
        for probe in probe_gene_count_dict:
            for geneName in probe_gene_count_dict[probe]:
                barcode_count = len(probe_gene_count_dict[probe][geneName])
                umi_count = 0
                read_count = 0
                for barcode in probe_gene_count_dict[probe][geneName]:
                    for umi in probe_gene_count_dict[probe][geneName][barcode]:
                        umi_count += len(
                            probe_gene_count_dict[probe][geneName][barcode]
                        )
                        read_count += probe_gene_count_dict[probe][geneName][barcode][
                            umi
                        ]
                row_list.append(
                    {
                        "probe": probe,
                        "gene": geneName,
                        "barcode_count": barcode_count,
                        "read_count": read_count,
                        "UMI_count": umi_count,
                    }
                )

        df_probe = pd.DataFrame(
            row_list,
            columns=["probe", "gene", "barcode_count", "read_count", "UMI_count"],
        )
        df_probe = df_probe.groupby(["probe"]).apply(
            lambda x: x.sort_values("UMI_count", ascending=False)
        )
        return df_probe

    def run(self):
        df_probe = self.bam2table()
        df_probe.to_csv(
            f"{self.outdir}/{self.sample}_probe_gene_count.tsv", sep="\t", index=False
        )

        df = pd.read_table(self.count_detail_file, header=0)
        # df_sum
        df_sum = Count.get_df_sum(df)

        # export all matrix
        self.write_matrix_10X(df, self.raw_matrix_dir)

        # call cells
        cell_bc, _threshold = self.cell_calling(df_sum)

        # get cell stats
        CB_describe = self.get_cell_stats(df_sum, cell_bc)

        # export cell matrix
        df_cell = df.loc[df["Barcode"].isin(cell_bc), :]
        self.write_matrix_10X(df_cell, self.cell_matrix_dir)
        (CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome) = (
            self.cell_summary(df, cell_bc)
        )

        # downsampling
        cell_bc = set(cell_bc)
        self.downsample(df_cell)

        # summary
        self.get_summary(
            CB_describe, CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome
        )

        self.report_prepare()

        self.add_content_item("metric", downsample_dict=self.downsample_dict)
        self._clean_up()


@utils.add_log
def count_capture_rna(args):
    # TODO!
    # need barcode_capture_rna
    with Count_capture_rna(args) as runner:
        runner.run()


def get_opts_count_capture_rna(parser, sub_program):
    get_opts_count(parser, sub_program)



import pandas as pd

from celescope.tools.step import Step, s_common
from celescope.tools.plotly_plot import Tsne_dropdown_plot, Tsne_single_plot, Tsne_plot


def get_opts_analysis_cite(parser, sub_program):
    if sub_program:
        parser.add_argument("--tsne_coord", help="tsne coord file", required=True)
        s_common(parser)


def analysis_cite(args):
    with Analysis_cite(args, display_title="Analysis") as runner:
        runner.run()


class Analysis_cite(Step):
    def __init__(self, args, display_title):
        super().__init__(args, display_title)

        # input_file
        self.tsne_coord = self.args.tsne_coord

    def run(self):
        df_tsne = pd.read_csv(self.tsne_coord, sep="\t", index_col=0)
        feature_name_list = [x[:45] for x in df_tsne.columns[4:]]
        df_tsne.columns = df_tsne.columns[:4].to_list() + feature_name_list

        # accelerate
        df_tsne["tSNE_1"] = df_tsne["tSNE_1"].astype("float16")
        df_tsne["tSNE_2"] = df_tsne["tSNE_2"].astype("float16")
        for col in feature_name_list:
            df_tsne[col] = df_tsne[col].astype("float16")

        # plot
        tsne_cluster = Tsne_plot(df_tsne, "cluster").get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)
        tsne_citeseq = Tsne_dropdown_plot(
            df_tsne, "Citeseq", feature_name_list
        ).get_plotly_div()
        self.add_data(tsne_citeseq=tsne_citeseq)
        Tsne_single_plot(df_tsne, feature_name_list, self.args.outdir).get_plotly_div()



__ASSAY__ = "capture_rna"

__STEPS__ = [
    "sample",
    "barcode",
    "cutadapt",
    "star",
    "featureCounts",
    "count_capture_rna",
    "analysis",
]

IMPORT_DICT = {
    "star": "celescope.rna",
    "analysis": "celescope.rna",
}



import pandas as pd

from celescope.tools.plotly_plot import Tsne_plot
from celescope.tools.step import Step
from celescope.tools.capture.__init__ import SUM_UMI_COLNAME
from celescope.tools import analysis_wrapper


def get_opts_analysis(parser, sub_program):
    if sub_program:
        parser.add_argument("--filter_umi_file", help="filter umi file", required=True)
        analysis_wrapper.get_opts_analysis_match(parser, sub_program)


class Analysis(Step):
    def __init__(self, args, display_title="Analysis"):
        super().__init__(args, display_title)
        self.df_filter_umi = pd.read_csv(args.filter_umi_file, index_col=0)

        # out
        self.df_tsne_file = f"{self.out_prefix}_UMI_tsne.csv"
        self.df_tsne = None

    def add_tsne_info(self):
        report_runner = analysis_wrapper.Report_runner(self.args)
        df_tsne, _df_marker = report_runner.get_df()
        self.df_tsne = pd.merge(
            df_tsne, self.df_filter_umi, left_index=True, right_index=True
        )
        self.df_tsne.to_csv(self.df_tsne_file)

    def add_cluster_metrics(self):
        self.add_help_content(
            "cluster 1,2,3...",
            "number of positive cells(sum_UMI > 0) in each cluster after filtering",
        )

        df_cluster_all = self.df_tsne.groupby("cluster").count()

        df_positive = self.df_tsne[self.df_tsne[SUM_UMI_COLNAME] > 0]
        df_cluster_positive = df_positive.groupby("cluster").count()

        for index, row in df_cluster_positive.iterrows():
            self.add_metric(
                name=f"cluster {index}",
                value=int(row["tSNE_1"]),
                total=int(df_cluster_all.loc[index, "tSNE_1"]),
            )

    def run(self):
        self.add_tsne_info()
        self.add_cluster_metrics()

        tsne_cluster = Tsne_plot(self.df_tsne, "cluster").get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)

        tsne_plot = Tsne_plot(self.df_tsne, SUM_UMI_COLNAME, discrete=False)
        tsne_plot.set_color_scale(["LightGrey", "Orange", "Red"])
        tsne_feature = tsne_plot.get_plotly_div()
        self.add_data(tsne_feature=tsne_feature)



import pandas as pd
import pysam

from collections import OrderedDict
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN
from celescope.flv_CR.summarize import Summarize
from celescope.tools import utils
from celescope.tools.step import s_common, Step
from celescope.tools.plotly_plot import Bar_plot


def gen_vj_annotation_metrics(df, seqtype):
    """
    Generate vdj Annotation Metrics from contig annotations file.
    """

    def get_vj_spanning_pair():
        """
        Get Productive V-J Spanning_Pair metric from annotation file
        Return productive chain pair number. eg: TRA/TRB or IGH/IGL, IGH/IGK.
        """
        df_productive = df[df["productive"] == True]

        if seqtype == "BCR":
            df_chain_heavy = df_productive[(df_productive["chain"] == "IGH")]
            df_chain_light = df_productive[
                (df_productive["chain"] == "IGL") | (df_productive["chain"] == "IGK")
            ]
        else:
            df_chain_heavy = df_productive[df_productive["chain"] == "TRA"]
            df_chain_light = df_productive[df_productive["chain"] == "TRB"]

        for _df in [df_chain_heavy, df_chain_light]:
            _df.drop_duplicates(["barcode"], inplace=True)

        VJ_Spanning_Pair_Cells = pd.merge(
            df_chain_heavy, df_chain_light, on="barcode", how="inner"
        )

        return VJ_Spanning_Pair_Cells.shape[0]

    metric_dict = OrderedDict()
    chains, chain_pairs = CHAIN[seqtype], PAIRED_CHAIN[seqtype]

    metric_dict["Cells match with scRNA-seq analysis"] = len(set(df.barcode))
    metric_dict["Cells With Productive V-J Spanning Pair"] = get_vj_spanning_pair()

    for pair in chain_pairs:
        chain1, chain2 = pair.split("_")[0], pair.split("_")[1]
        cbs1 = set(df[(df["productive"] == True) & (df["chain"] == chain1)].barcode)
        cbs2 = set(df[(df["productive"] == True) & (df["chain"] == chain2)].barcode)
        metric_dict[f"Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair"] = (
            len(cbs1.intersection(cbs2))
        )

    for chain in chains:
        metric_dict[f"Cells With {chain} Contig"] = len(
            set(df[df["chain"] == chain].barcode)
        )
        metric_dict[f"Cells With CDR3-annotated {chain} Contig"] = len(
            set(df[(df["chain"] == chain) & (df["cdr3"] != "None")].barcode)
        )
        metric_dict[f"Cells With V-J Spanning {chain} Contig"] = len(
            set(df[(df["full_length"] == True) & (df["chain"] == chain)].barcode)
        )
        metric_dict[f"Cells With Productive {chain} Contig"] = len(
            set(df[(df["productive"] == True) & (df["chain"] == chain)].barcode)
        )

    return metric_dict


def gen_clonotypes_table(df, out_clonotypes):
    """
    Generate clonotypes.csv file
    """
    df = df[df["productive"] == True]
    df["chain_cdr3aa"] = df[["chain", "cdr3"]].apply(":".join, axis=1)

    match_clonotypes = open(out_clonotypes, "w")
    match_clonotypes.write("barcode\tcdr3s_aa\n")
    for cb in set(df.barcode):
        temp = df[df["barcode"] == cb].sort_values(by="chain", ascending=True)
        chain_pair = ";".join(temp["chain_cdr3aa"].tolist())
        match_clonotypes.write(f"{cb}\t{chain_pair}\n")
    match_clonotypes.close()

    df_clonetypes = pd.read_csv(out_clonotypes, sep="\t", index_col=None)
    df_clonetypes = df_clonetypes.groupby("cdr3s_aa", as_index=False).agg(
        {"barcode": "count"}
    )
    df_clonetypes.rename(columns={"barcode": "frequency"}, inplace=True)
    sum_f = df_clonetypes["frequency"].sum()
    df_clonetypes["proportion"] = df_clonetypes["frequency"].apply(lambda x: x / sum_f)
    df_clonetypes["clonotype_id"] = [
        f"clonotype{i}" for i in range(1, df_clonetypes.shape[0] + 1)
    ]
    df_clonetypes = df_clonetypes.reindex(
        columns=["clonotype_id", "cdr3s_aa", "frequency", "proportion"]
    )
    df_clonetypes.sort_values(by="frequency", ascending=False, inplace=True)
    df_clonetypes.to_csv(out_clonotypes, sep=",", index=False)


class Match(Step):
    """
    ## Features

    - Assembled results match with sc-RNA library.

    - Generate matched VDJ-annotation metrics, clonetypes table and bar-plot of clonetypes distribution in html.

    ## Output

    - `matched_contig_annotations.csv` High-level annotations of each high-confidence contigs from matched cell-associated barcodes.

    - `matched_contig.fasta` High-confidence contig sequences annotated in the matched_contig_annotations.csv.

    - `matched_productive_contig_annotations.csv` Annotations of each productive contigs from matched cell-associated barcodes. This is a subset of matched_contig_annotations.csv.

    - `matched_productive_contig.fasta` Productive contig sequences annotated in the matched_productive_contig_annotations.csv.

    - `clonotypes.csv` High-level descriptions of each clonotype where barcodes match with scRNA-Seq.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.chains = CHAIN[self.seqtype]
        self.pairs = PAIRED_CHAIN[self.seqtype]

        self.match_dir = args.match_dir
        if self.match_dir != "None":
            self.match_cell_barcodes, _ = utils.get_barcode_from_match_dir(
                self.match_dir
            )

        self.filter_annotation = f"{args.summarize_out}/filtered_contig_annotations.csv"
        self.filter_fasta = f"{args.summarize_out}/filtered_contig.fasta"
        self.clonotypes = f"{args.summarize_out}/clonotypes.csv"

        # out
        self.match_annotation = f"{self.outdir}/matched_contig_annotations.csv"
        self.match_fasta = f"{self.outdir}/matched_contig.fasta"
        self.match_clonotypes = f"{self.outdir}/matched_clonotypes.csv"

    @utils.add_log
    def gen_matched_result(self):
        """
        Generate annotation and fasta files where barcodes matched with scRNA.
        """
        SGR_annotation_file = pd.read_csv(self.filter_annotation)
        match_annotation_file = SGR_annotation_file[
            SGR_annotation_file.barcode.isin(self.match_cell_barcodes)
        ]
        match_annotation_file.to_csv(self.match_annotation, sep=",", index=False)

        SGR_fasta_file = pysam.FastxFile(self.filter_fasta)
        match_fasta_file = open(self.match_fasta, "w")
        for entry in SGR_fasta_file:
            name = entry.name
            attrs = name.split("_")
            cb = attrs[0]
            if cb in self.match_cell_barcodes:
                new_name = cb + "_" + attrs[1] + "_" + attrs[2]
                seq = entry.sequence
                match_fasta_file.write(f">{new_name}\n{seq}\n")
        match_fasta_file.close()

        Summarize.gen_productive_contig(
            match_annotation_file, self.match_fasta, self.outdir, prefix="matched_"
        )

    @utils.add_log
    def gen_matched_clonotypes(self):
        """
        Generate clonotypes.csv file where barcodes match with scRNA
        """
        df_match = pd.read_csv(self.match_annotation)
        gen_clonotypes_table(df_match, self.match_clonotypes)

    @utils.add_log
    def gen_matched_metrics(self):
        """
        Generate Matched vdj Annotation Metrics.
        """
        df_match = pd.read_csv(self.match_annotation)
        metrics_dict = gen_vj_annotation_metrics(df_match, self.seqtype)

        for k, v in metrics_dict.items():
            if k == "Cells match with scRNA-seq analysis":
                self.add_metric(
                    name=k,
                    value=v,
                    help_info="The intersection between VDJ cell barcodes and scRNA-Seq barcodes. All the following metrics are based on this intersection.",
                )
            else:
                self.add_metric(name=k, value=v, total=len(set(df_match.barcode)))

    @utils.add_log
    def gen_clonotypes_table(self):
        title = "Clonetypes"
        raw_clonotypes = pd.read_csv(self.clonotypes, sep=",", index_col=None)
        raw_clonotypes["ClonotypeID"] = raw_clonotypes["clonotype_id"].apply(
            lambda x: x.strip("clonetype")
        )
        raw_clonotypes["Frequency"] = raw_clonotypes["frequency"]
        raw_clonotypes["Proportion"] = raw_clonotypes["proportion"].apply(
            lambda x: f"{round(x*100, 2)}%"
        )
        raw_clonotypes["CDR3_aa"] = raw_clonotypes["cdr3s_aa"].apply(
            lambda x: x.replace(";", "<br>")
        )

        table_dict = self.get_table_dict(
            title=title,
            table_id="clonetypes",
            df_table=raw_clonotypes[
                ["ClonotypeID", "CDR3_aa", "Frequency", "Proportion"]
            ],
        )
        self.add_data(table_dict=table_dict)

        raw_clonotypes["ClonotypeID"] = raw_clonotypes["ClonotypeID"].astype("int")
        raw_clonotypes.sort_values(by=["ClonotypeID"], inplace=True)
        Barplot = Bar_plot(df_bar=raw_clonotypes).get_plotly_div()
        self.add_data(Barplot=Barplot)

    @utils.add_log
    def run(self):
        if self.match_dir != "None":
            self.gen_matched_result()
            self.gen_matched_clonotypes()
            self.gen_matched_metrics()
        self.gen_clonotypes_table()


def match(args):
    with Match(args, display_title="Match") as runner:
        runner.run()


def get_opts_match(parser, sub_program):
    parser.add_argument(
        "--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True
    )
    if sub_program:
        s_common(parser)
        parser.add_argument(
            "--match_dir", help="scRNA-seq match directory", required=True
        )
        parser.add_argument(
            "--summarize_out",
            help="assemble result in SGR barcode from summarize directory",
            required=True,
        )
    return parser



import re
import subprocess

import pysam

from celescope.tools.step import Step, s_common
from celescope.tools import utils

ADAPTER = ["polyT=A{18}", "p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC"]

LOG_METRICS_TITLE = (
    "Total reads processed",
    "Reads with adapters",
    "Reads that were too short",
    "Reads written (passing filters)",
    "Total basepairs processed",
    "Quality-trimmed",
    "Total written (filtered)",
)


def read_cutadapt_log(cutadapt_log):
    """
    Returns:
        dict: {
            'Total reads processed': int, 'Reads with adapters': int, 'Reads that were too short': int,
            'Reads written (passing filters)': int, 'Total basepairs processed': int,
            'Quality-trimmed': int, 'Total written (filtered)': int
        }
    """
    metrics_dict = {}
    remove_strs = [r",", r" bp", r"\(.*\)"]

    for _line_index, line in enumerate(cutadapt_log.split("\n")):
        line = line.strip()
        if not line:
            continue
        attr = line.split(":")
        if attr[0] in LOG_METRICS_TITLE:
            title, number = attr[0], attr[1]
            number = attr[1].strip()
            for remove_str in remove_strs:
                number = re.sub(pattern=remove_str, repl="", string=number)
            metrics_dict[title] = int(number)

    return metrics_dict


class Cutadapt(Step):
    """
    ## Features
    - Trim adapters in R2 reads with cutadapt. Default adapters includes:
        - polyT=A{18}, 18 A bases.
        - p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCA, Illumina p5 adapter.
    ## Output
    - `cutadapt.log` Cutadapt output log file.
    - `{sample}_clean_2.fq.gz` R2 reads file without adapters.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        # set
        self.adapter_args = self.read_adapter_fasta(args.adapter_fasta)
        self.adapter_args += ADAPTER

        # out files
        if args.gzip:
            suffix = ".gz"
        else:
            suffix = ""
        self.out_fq2 = f"{self.outdir}/{self.sample}_clean_2.fq{suffix}"
        self.cutadapt_log_file = f"{self.outdir}/cutadapt.log"

    @staticmethod
    def read_adapter_fasta(adapter_fasta):
        """
        return ['adapter1=AAA','adapter2=BBB']
        """
        adapter_args = []
        if adapter_fasta and adapter_fasta != "None":
            with pysam.FastxFile(adapter_fasta) as fh:
                for read in fh:
                    adapter_args.append(f"{read.name}={read.sequence}")
        return adapter_args

    def add_cutadapt_metrics(self, cutadapt_log):
        # Total reads processed:...Total written (filtered):
        metrics_dict = read_cutadapt_log(cutadapt_log)

        total_reads = metrics_dict["Total reads processed"]
        reads_with_adapters = metrics_dict["Reads with adapters"]
        reads_too_short = metrics_dict["Reads that were too short"]
        reads_written = metrics_dict["Reads written (passing filters)"]
        total_base_pairs = metrics_dict["Total basepairs processed"]
        quality_trimmed = metrics_dict["Quality-trimmed"]
        base_pairs_written = metrics_dict["Total written (filtered)"]

        self.add_metric(
            name="Reads with Adapters",
            value=reads_with_adapters,
            total=total_reads,
            help_info="R2 reads with poly A(read-through adpaters) or sequencing adapters",
        )
        self.add_metric(
            name="Reads too Short",
            value=reads_too_short,
            total=total_reads,
            help_info="reads with read length less than 20bp after trimming",
        )
        self.add_metric(
            name="Reads Written",
            value=reads_written,
            total=total_reads,
            help_info="reads pass filtering",
        )
        self.add_metric(
            name="Base Pairs Processed",
            value=total_base_pairs,
            help_info="total processed base pairs",
        )
        self.add_metric(
            name="Base Pairs Quality-Trimmed",
            value=quality_trimmed,
            total=total_base_pairs,
            help_info="bases pairs removed from the end of the read whose quality is smaller than the given threshold",
        )
        self.add_metric(
            name="Base Pairs Written",
            value=base_pairs_written,
            total=total_base_pairs,
            help_info="base pairs pass filtering",
        )

    @utils.add_log
    def run(self):
        adapter_args_str = " ".join(["-a " + adapter for adapter in self.adapter_args])
        cmd = (
            "cutadapt "
            f"{adapter_args_str} "
            f"-n {len(self.adapter_args)} "
            f"-j {self.thread} "
            f"-m {self.args.minimum_length} "
            f"--nextseq-trim={self.args.nextseq_trim} "
            f"--overlap {self.args.overlap} "
            f"-l {self.args.insert} "
            f"-o {self.out_fq2} "
            f"{self.args.cutadapt_param} "
            f"{self.args.fq} "
        )
        self.run.logger.info(cmd)
        # need encoding argument to return str
        results = subprocess.run(
            cmd, stdout=subprocess.PIPE, encoding="utf-8", check=True, shell=True
        )
        cutadapt_log = results.stdout
        with open(self.cutadapt_log_file, "w") as f:
            f.write(cutadapt_log)
        self.add_cutadapt_metrics(cutadapt_log)


@utils.add_log
def cutadapt(args):
    with Cutadapt(args, display_title="Trimming") as runner:
        runner.run()


def get_opts_cutadapt(parser, sub_program):
    parser.add_argument(
        "--gzip", help="Output gzipped fastq files.", action="store_true"
    )
    parser.add_argument("--adapter_fasta", help="Addtional adapter fasta file.")
    parser.add_argument(
        "--minimum_length",
        help="Discard processed reads that are shorter than LENGTH.",
        default=20,
    )
    parser.add_argument(
        "--nextseq_trim",
        help="""Quality trimming of reads using two-color chemistry (NextSeq). 
Some Illumina instruments use a two-color chemistry to encode the four bases. 
This includes the NextSeq and the NovaSeq. 
In those instruments, a ‘dark cycle’ (with no detected color) encodes a G. 
However, dark cycles also occur when sequencing “falls off” the end of the fragment.
The read then contains a run of high-quality, but incorrect “G” calls at its 3’ end.""",
        default=20,
    )
    parser.add_argument(
        "--overlap",
        help="""Since Cutadapt allows partial matches between the read and the adapter sequence,
short matches can occur by chance, leading to erroneously trimmed bases. 
For example, roughly 0.25 of all reads end with a base that is identical to the first base of the adapter. 
To reduce the number of falsely trimmed bases, the alignment algorithm requires that 
at least {overlap} bases match between adapter and read. """,
        default=10,
    )
    parser.add_argument("--insert", help="Read2 insert length.", default=150)
    parser.add_argument(
        "--cutadapt_param",
        help='Other cutadapt parameters. For example, --cutadapt_param "-g AAA" ',
        default="",
    )
    if sub_program:
        parser.add_argument(
            "--fq", help="Required. R2 reads from step Barcode.", required=True
        )
        parser = s_common(parser)
    return parser



__STEPS__ = [
    "mkref",
    "sample",
    "barcode",
    "cutadapt",
    "star",
    "featureCounts",
    "count",
    "analysis",
]
__ASSAY__ = "rna"

# m: memory
# x: thread
RESOURCE = {
    "sample": {"m": 1, "x": 1},
    "barcode": {"m": 5, "x": 1},
    "cutadapt": {"m": 5, "x": 1},
    "star": {"m": 30, "x": 1},
}



__STEPS__ = [
    "sample",
    "barcode",
    "cutadapt",
    "star",
    "star_virus",
    "featureCounts",
    "count",
    "count_virus",
    "analysis_rna_virus",
]
__ASSAY__ = "rna_virus"
IMPORT_DICT = {"star": "celescope.rna"}



import numpy as np
import pandas as pd
import scipy.io
import scipy.sparse
import glob
import scanpy as sc
import anndata as ad

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.matrix import CountMatrix, Features


TAG_COL = "tag_name"


def filter_fun(df):
    """
    Filter data according to mean and standard deviation
    """
    df["filter_num"] = np.mean(df, axis=1) + np.std(df, axis=1)
    for index, _ in df.iterrows():
        df.loc[index] = df.loc[index].mask(
            df.loc[index] < df.loc[index]["filter_num"], 0
        )
    df.drop(columns="filter_num", inplace=True)
    return df


class CountMatrix2(CountMatrix):
    @classmethod
    def from_dataframe2(cls, df, features: Features, barcodes=None):
        matrix = df
        mtx = scipy.sparse.coo_matrix(matrix, matrix.shape, int)
        return cls(features, barcodes, mtx)


class Count_cite(Step):
    def __init__(self, args, display_title):
        super().__init__(args, display_title)

        self.df_read_count = pd.read_csv(args.read_count_file, sep="\t", index_col=0)

        self.match_dict = utils.parse_match_dir(args.match_dir)
        self.match_barcode = self.match_dict["match_barcode"]
        self.match_matrix_dir = self.match_dict["matrix_dir"]

        # input
        self.tsne_coord = glob.glob(f"{args.match_dir}/*analysis*/*tsne_coord.tsv")[0]

        # out
        self.mtx = f"{self.out_prefix}_citeseq.mtx.gz"
        self.raw_matrix_dir = f"{self.out_prefix}_raw_citeseq_matrix"
        self.filtered_matrix_dir = f"{self.out_prefix}_filtered_citeseq_matrix"
        self.filtered_tsne_coord = f"{self.out_prefix}_filtered_tsne_coord.tsv"

    @utils.add_log
    def run(self):
        mapped_read = int(self.df_read_count["read_count"].sum())

        # in cell
        df_read_count_in_cell = self.df_read_count[
            self.df_read_count.index.isin(self.match_barcode)
        ]
        mapped_read_in_cell = int(df_read_count_in_cell["read_count"].sum())
        self.add_metric(
            name="Mapped Reads in Cells",
            value=mapped_read_in_cell,
            total=mapped_read,
        )

        tag_names = df_read_count_in_cell[TAG_COL].unique()
        features_raw = Features(tag_names)

        # raw_matrix
        raw_citeseq_matrix = CountMatrix.from_dataframe(
            df_read_count_in_cell,
            features_raw,
            barcodes=self.match_barcode,
            row=TAG_COL,
            column="barcode",
            value="UMI",
        )
        rna_matrix = CountMatrix.from_matrix_dir(matrix_dir=self.match_matrix_dir)
        raw_merged_matrix = rna_matrix.concat_by_barcodes(raw_citeseq_matrix)
        raw_merged_matrix.to_matrix_dir(self.raw_matrix_dir)

        # UMI
        df_UMI_in_cell = (
            df_read_count_in_cell.reset_index()
            .groupby(["barcode", TAG_COL])
            .agg({"UMI": "count"})
        )

        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index="barcode", columns=TAG_COL, values="UMI"
        )
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell, df_UMI_in_cell, how="left", left_index=True, right_index=True
        )

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        df_UMI_cell = df_UMI_cell.astype(int)
        df_UMI_cell_out = df_UMI_cell.T
        df_UMI_cell_out.to_csv(self.mtx, sep="\t", compression="gzip")

        # fix
        ## 1.normalize;2.filter
        ### normalize
        obs = pd.DataFrame(index=df_UMI_cell_out.columns)
        var = pd.DataFrame(
            df_UMI_cell_out.index, index=df_UMI_cell_out.index, columns=["ADT"]
        )
        mdata_citeseq = ad.AnnData(np.array(df_UMI_cell_out.T), obs=obs, var=var)
        sc.pp.normalize_total(
            mdata_citeseq,
            target_sum=1e4,
            inplace=True,
        )
        sc.pp.log1p(mdata_citeseq)
        ### filter
        df_filtered = filter_fun(mdata_citeseq.to_df().T)
        features_filtered = Features(df_filtered.index.to_list())
        filtered_citeseq_matrix = CountMatrix2.from_dataframe2(
            df_filtered, features_filtered, barcodes=self.match_barcode
        )
        filtered_merged_matrix = rna_matrix.concat_by_barcodes(filtered_citeseq_matrix)
        filtered_merged_matrix.to_matrix_dir(self.filtered_matrix_dir)

        # filtered_tsne.csv
        df_tsne = pd.read_csv(self.tsne_coord, sep="\t")
        if "Unnamed: 0" in df_tsne.columns:
            df_tsne.rename(columns={"Unnamed: 0": "barcode"}, inplace=True)
            df_tsne = df_tsne.set_index("barcode")

        df_citeseq = mdata_citeseq.to_df()
        df_tsne = pd.concat([df_tsne, df_citeseq], axis=1)
        df_tsne.fillna(0, inplace=True)
        df_tsne.to_csv(self.filtered_tsne_coord, sep="\t")

        # UMI
        UMIs = df_UMI_cell.apply(sum, axis=1)
        median_umi = round(np.median(UMIs), 2)
        mean_umi = round(np.mean(UMIs), 2)
        self.add_metric(
            name="Median UMI per Cell",
            value=float(median_umi),
        )
        self.add_metric(
            name="Mean UMI per Cell",
            value=float(mean_umi),
        )


def get_opts_count_cite(parser, sub_program):
    if sub_program:
        parser.add_argument("--match_dir", help=HELP_DICT["match_dir"], required=True)
        parser.add_argument("--read_count_file", help="tag read count file")
        s_common(parser)


def count_cite(args):
    with Count_cite(args, display_title="Count") as runner:
        runner.run()



from celescope.tools import utils
from celescope.__init__ import __VERSION__
from celescope.tools.__init__ import PATTERN_DICT
from celescope.tools.barcode import Chemistry
from celescope.tools.step import Step, s_common


def add_kit_version(chemistry):
    kit_dict = {
        "1": "no longer in use",
        "2": "kit V1",
        "3": "kit V2",
    }
    if chemistry.startswith("scopeV"):
        s = chemistry.replace("scopeV", "")
        chem_version = s[0]
        kit = kit_dict[chem_version]
        chemistry = f"{chemistry} ({kit})"

    return chemistry


class Sample(Step):
    def __init__(self, args):
        Step.__init__(self, args)
        self.assay_description = utils.get_assay_text(self.assay)
        self.version = __VERSION__
        self.chemistry = args.chemistry

    @utils.add_log
    def run(self):
        if self.chemistry == "auto":
            fq1 = self.args.fq1
            ch = Chemistry(fq1)
            chemistry = ch.check_chemistry()
            chemistry = ",".join(set(chemistry))
        else:
            chemistry = self.chemistry

        self.add_metric(
            name="Sample ID",
            value=self.sample,
        )
        self.add_metric(
            name="Assay",
            value=self.assay_description,
        )
        self.add_metric(
            name="Chemistry",
            value=chemistry,
            display=add_kit_version(chemistry),
            help_info='For more information, see <a href="https://github.com/singleron-RD/CeleScope/blob/master/docs/chemistry.md">here</a>',
        )
        self.add_metric(
            name="Software Version",
            value=self.version,
        )


@utils.add_log
def sample(args):
    with Sample(args) as runner:
        runner.run()


def get_opts_sample(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--fq1", help="read1 fq file")
    parser.add_argument(
        "--chemistry",
        choices=list(PATTERN_DICT.keys()),
        help="chemistry version",
        default="auto",
    )
    return parser



import json
import pandas as pd
import pysam
import os
import subprocess

from celescope.tools import utils
from celescope.tools.step import s_common, Step


class Summarize(Step):
    """
    ## Features

    - Convert 10X barcode of assemble result back to SGR barcode.

    - Generate Productive contigs sequences and annotation files.

    - Generate VDJ-annotation metrics in html.

    ## Output

    - `filtered_contig_annotations.csv` High-level annotations of each high-confidence contigs from cell-associated barcodes.

    - `filtered_contig.fasta` High-confidence contig sequences annotated in the filtered_contig_annotations.csv.

    - `productive_contig_annotations.csv` Annotations of each productive contigs from cell-associated barcodes. This is a subset of filtered_contig_annotations.csv.

    - `productive_contig.fasta` Productive contig sequences annotated in the productive_contig_annotations.csv.

    - `clonotypes.csv` High-level descriptions of each clonotype.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.version = os.path.dirname(args.soft_path).split("/")[-1].split("-")[-1]

        with open(args.barcode_convert_json, "r") as f:
            self.tenX_sgr = json.load(f)

        # assemble result directory in 10X barcode.
        self.assemble_out = args.assemble_out

        # in files
        annotation_file = f"{self.assemble_out}/filtered_contig_annotations.csv"
        self.df_annotation = pd.read_csv(annotation_file, sep=",", index_col=None)
        tenX_fasta_file = f"{self.assemble_out}/filtered_contig.fasta"
        self.tenX_fasta_fh = pysam.FastxFile(tenX_fasta_file)

        # out
        self.filtered_contig_annotations = (
            f"{self.outdir}/filtered_contig_annotations.csv"
        )
        self.filter_contig_fasta = f"{self.outdir}/filtered_contig.fasta"

    @utils.add_log
    def convert_barcode_to_SGR(self):
        """
        Convert 10X barcode to SGR barcode format.
        """

        self.df_annotation["barcode"] = self.df_annotation["barcode"].apply(
            lambda x: self.tenX_sgr[x.split("-")[0]]
        )
        self.df_annotation["contig_id"] = self.df_annotation["contig_id"].apply(
            lambda x: self.tenX_sgr[x.split("-")[0]]
            + "_"
            + x.split("_")[1]
            + "_"
            + x.split("_")[2]
        )
        if int(self.version.split(".")[0]) < 4:
            self.df_annotation["productive"].replace(
                {"True": True, "None": False}, inplace=True
            )

        self.df_annotation.to_csv(
            self.filtered_contig_annotations, sep=",", index=False
        )

        with open(self.filter_contig_fasta, "w") as f:
            for entry in self.tenX_fasta_fh:
                name = entry.name
                seq = entry.sequence
                attrs = name.split("_")
                new_name = (
                    self.tenX_sgr[attrs[0].split("-")[0]]
                    + "_"
                    + attrs[1]
                    + "_"
                    + attrs[2]
                )
                f.write(f">{new_name}\n{seq}\n")

        cmd = f"cp {self.assemble_out}/clonotypes.csv {self.outdir}"
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    @utils.add_log
    def gen_productive_contig(annotation_file, fasta_file, outdir, prefix=""):
        """
        Generate productive_contig_annotation.csv and productive_contig.fasta file.
        """
        productive_contig = annotation_file[annotation_file["productive"] == True]
        productive_contig_id = set(productive_contig["contig_id"])
        productive_contig.to_csv(
            f"{outdir}/{prefix}productive_contig_annotations.csv", sep=",", index=False
        )

        productive_fasta = open(f"{outdir}/{prefix}productive_contig.fasta", "w")
        with pysam.FastxFile(fasta_file, "r") as f:
            for read in f:
                if read.name in productive_contig_id:
                    productive_fasta.write(
                        ">" + read.name + "\n" + read.sequence + "\n"
                    )
        productive_fasta.close()

    def run(self):
        self.convert_barcode_to_SGR()
        Summarize.gen_productive_contig(
            self.df_annotation, self.filter_contig_fasta, self.outdir
        )


def summarize(args):
    with Summarize(args) as runner:
        runner.run()


def get_opts_summarize(parser, sub_program):
    parser.add_argument(
        "--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True
    )
    parser.add_argument("--soft_path", help="soft path for cellranger")
    if sub_program:
        s_common(parser)
        parser.add_argument("--barcode_convert_json", help="json file", required=True)
        parser.add_argument(
            "--assemble_out",
            help="directory of cellranger assemble result",
            required=True,
        )
    return parser
